# ================================================== #
#                  TEST COMPONENTS                   #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/21/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

import didery.static.transcrypt.components.searcher as searcher
import didery.static.transcrypt.components.tables as tables
import didery.static.transcrypt.components.tabs as tabs
import didery.static.transcrypt.router as router
from tester import test

o = require("mithril/ospec/ospec")
sinon = require("sinon")

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

@test
class Searcher:
    """
    Class for testing searcher.
    """
    class CaseSensitive:
        """
        Class for testing search case sensitivity.
        """
        def beforeEach(self):
            """
            Sets up searcher.
            """
            self.searcher = searcher.Searcher()

        # ========================================== #

        def setSearch(self):
            """
            Tests case sensitivity.
            """
            self.searcher.setSearch('"SensITive"')
            o(self.searcher.searchTerm).equals("SensITive")("Checking search term.")
            o(self.searcher.caseSensitive).equals(True)("Checking case sensitivity.")

        # ========================================== #

        def searchBasic(self):
            """
            Tests basic search.
            """
            self.searcher.setSearch('"Foo"')

            o(self.searcher.search({
                "bar": "foo"
            })).equals(False)("Checking first basic search.")
            o(self.searcher.search({
                "foo": "bar"
            })).equals(False)("Checking second basic search.")
            o(self.searcher.search({
                "Foo": "bar"
            })).equals(False)("Checking third basic search.")
            o(self.searcher.search({
                "Bar": "Foo"
            })).equals(True)("Checking fourth basic search.")

        # ========================================== #

        def searchNested(self):
            """
            Tests nested search.
            """
            self.searcher.setSearch('"Foo"')

            o(self.searcher.search({
                "bar": {
                    "bar": "foo"
                }
            })).equals(False)("Checking first nested search.")
            o(self.searcher.search({
                "bar": {
                    "bar": "Foo"
                }
            })).equals(True)("Checking second nested search.")

        # ========================================== #

        def searchList(self):
            """
            Tests list search.
            """
            self.searcher.setSearch('"Foo"')

            o(self.searcher.search({
                "foo": [0, 1, "foo", 3]
            })).equals(False)("Checking first list search.")
            o(self.searcher.search({
                "foo": [0, 1, "Foo", 3]
            })).equals(True)("Checking second list search.")

        # ========================================== #

        def searchListNested(self):
            """
            Tests nested list search.
            """
            self.searcher.setSearch('"Foo"')

            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "foo"
                }, [3, "foo", 5], 6]
            })).equals(False)("Checking first nested list search.")
            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "foo"
                }, [3, "Foo", 5], 6]
            })).equals(True)("Checking second nested list search.")
            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "Foo"
                }, [3, "foo", 5], 6]
            })).equals(True)("Checking third nested list search.")

        # ========================================== #

        def searchInnerString(self):
            """
            Tests inner string search.
            """
            self.searcher.setSearch('"Foo"')

            o(self.searcher.search({
                "foo": "blah Fblahooblah blah"
            })).equals(False)("Checking first inner string search.")
            o(self.searcher.search({
                "foo": "blah blahFooblah blah"
            })).equals(True)("Checking second inner string search.")

    # ============================================== #

    class CaseInsensitive:
        """
        Class for testing search case sensitivity.
        """
        def beforeEach(self):
            """
            Sets up searcher.
            """
            self.searcher = searcher.Searcher()

        # ========================================== #

        def setSearch(self):
            """
            Tests case sensitivity.
            """
            self.searcher.setSearch("InSensItive")
            o(self.searcher.searchTerm).equals("insensitive")("Checking search term.")
            o(self.searcher.caseSensitive).equals(False)("Checking case sensitivity.")

        # ========================================== #

        def searchBasic(self):
            """
            Tests basic search.
            """
            self.searcher.setSearch('Foo')

            o(self.searcher.search({
                "bar": "foo"
            })).equals(True)("Checking first basic search.")
            o(self.searcher.search({
                "foo": "bar"
            })).equals(False)("Checking second basic search.")
            o(self.searcher.search({
                "Foo": "bar"
            })).equals(False)("Checking third basic search.")
            o(self.searcher.search({
                "Bar": "Foo"
            })).equals(True)("Checking fourth basic search.")

        # ========================================== #

        def searchNested(self):
            """
            Tests nested search.
            """
            self.searcher.setSearch('foo')

            o(self.searcher.search({
                "bar": {
                    "bar": "bar"
                }
            })).equals(False)("Checking first nested search.")
            o(self.searcher.search({
                "bar": {
                    "bar": "Foo"
                }
            })).equals(True)("Checking second nested search.")

        # ========================================== #

        def searchList(self):
            """
            Tests list search.
            """
            self.searcher.setSearch('foo')

            o(self.searcher.search({
                "foo": [0, 1, "bar", 3]
            })).equals(False)("Checking first list search.")
            o(self.searcher.search({
                "foo": [0, 1, "Foo", 3]
            })).equals(True)("Checking second list search.")

        # ========================================== #

        def searchInnerString(self):
            """
            Tests inner string search.
            """
            self.searcher.setSearch('foo')

            o(self.searcher.search({
                "foo": "blah Fblahooblah blah"
            })).equals(False)("Checking first inner string search.")
            o(self.searcher.search({
                "foo": "blah blahFooblah blah"
            })).equals(True)("Checking second inner string search.")

        # ========================================== #

        def searchListNested(self):
            """
            Tests nested list search.
            """
            self.searcher.setSearch('foo')

            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "boo"
                }, [3, "boo", 5], 6]
            })).equals(False)("Checking first nested list search.")
            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "Foo"
                }, [3, "boo", 5], 6]
            })).equals(True)("Checking second nested list search.")
            o(self.searcher.search({
                "bar": [0, 1, {
                    "bar": "boo"
                }, [3, "Foo", 5], 6]
            })).equals(True)("Checking third nested list search.")

# ================================================== #

@test
class TabledTab:
    """
    Class for testing tabs and tab content.
    """
    _Errors_Data = [{
                    "title": "Invalid Signature.",
                    "msg": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= had an invalid rotation signature.",
                    "time": "2000-01-01T00:00:00+00:00"
                },
                {
                    "title": "Relay Unreachable.",
                    "msg": "Could not establish a connection with relay servers.",
                    "time": "2000-01-01T11:00:00+00:00"
                }]
    _History_Data = [{
                "history":
                {
                    "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                    "changed" : "2001-01-01T00:00:00+00:00",
                    "signer": 2,
                    "signers":
                    [
                        "Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA="
                    ]
                },
                "signatures":
                [
                    "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCg==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            }, {
                "history":
                {
                    "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                    "changed" : "2000-01-01T00:00:00+00:00",
                    "signer": 1,
                    "signers":
                    [
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY="
                    ]
                },
                "signatures":
                [
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            }]
    _Blobs_Data = [{
                "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            },
            {
                "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            }]
    _Relays_Data = [{
                "host address": "127.0.0.1",
                "port": 7541,
                "name": "alpha",
                "main": True,
                "uid": "1",
                "status": "connected"
            },
            {
                "host address": "127.0.0.1",
                "port": 7542,
                "name": "beta",
                "main": False,
                "uid": "2",
                "status": "connected"
            }]

    def before(self):
        """
        Sets up fake server.
        """
        self._testServer = sinon.createFakeServer()
        self._testServer.respondWith("[]")
        self._testServer.respondImmediately = True
        window.XMLHttpRequest = XMLHttpRequest

    # ============================================== #

    def after(self):
        """
        Resets fake server.
        """
        self._testServer.restore()

    # ============================================== #

    def asyncBeforeEach(self, done):
        """
        Sets up tabs and pages.

            Parameters:
            done - Automatically included promise done function
        """
        r = router.Router()
        r.route()
        self.tabs = r.tabs
        setTimeout(done)

    # ============================================== #

    def startup(self):
        """
        Tests tab number and order.
        """
        self.tabs.searchAll()
        o(self.tabs.searcher.searchTerm).equals("")("Checking default search string.")
        o(len(self.tabs.tabs)).equals(4)("Checking tab number.")

        current = self.tabs.currentTab()
        o(current.DataTab).equals(self.tabs.tabs[0].DataTab)("Checking first tab.")
        o(type(current)).equals(tabs.History)("Checking tab order.")

        o(self.tabs.tabs[1].DataTab).equals(tabs.Blobs.DataTab)("Checking second tab.")
        o(self.tabs.tabs[2].DataTab).equals(tabs.Relays.DataTab)("Checking third tab.")
        o(self.tabs.tabs[3].DataTab).equals(tabs.Errors.DataTab)("Checking fourth tab.")

    # ============================================== #

    def _setData(self, callback):
        """
        Sets tab data.

            Parameters:
            callback - Callback function
        """
        self.tabs.currentTab().table._setData(self._History_Data)
        self._redraw(callback)

    # ============================================== #

    def _redraw(self, callback):
        """
        Redraws document.

            Parameters:
            callback - Callback function
        """
        m.redraw()
        setTimeout(callback, 50)

    # ============================================== #

    def _clickRow(self, row, callback):
        """
        Clicks row.

            Parameters:
            row - Row index
            callback - Callback function
        """
        self.tabs.tabs[0]._getRows()[row].click()
        self._redraw(callback)

    # ============================================== #

    def _clickId(self, id, callback):
        """
        Clicks element with given id.

            Parameters:
            id - Element id
            callback - Callback function
        """
        jQuery("#" + id).click()
        self._redraw(callback)

    # ============================================== #

    def _tableIsEmpty(self, rows):
        """
        Tests if table is empty.

            Parameters:
            rows - Table rows
        """
        o(rows.length).equals(1)("Checking row length.")
        td = rows.find("td")
        o(td.length).equals(1)("Checking td length.")
        o(td.text()).equals(tables.Table.no_results_text)("Checking table content.")

    # ============================================== #

    def asyncBasicSearch(self, done, timeout):
        """
        Tests basic search.

            Parameters:
            done - Automatically included promise done function
            timeout - Automatically included timeout
        """
        timeout(200)

        jQuery("#" + self.tabs._searchId).val("did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=")
        self.tabs.searchAll()
        o(self.tabs.searcher.searchTerm).equals("did:dad:qt27fthwonzsa88vrtkep6h-4ha8tr54shon1vwl6fe=")("Checking search query.")

        def f1():
            history = self.tabs.tabs[0]
            o(history._getRows().length).equals(1)("Checking visible rows.")
            o(history._getLabel().text()).equals("1/2")("Checking label text.")

            relays = self.tabs.tabs[2]
            self._tableIsEmpty(relays._getRows())
            o(relays._getLabel().text()).equals("0/2")("Checking label text.")

            def f2():
                self._tableIsEmpty(history._getRows())
                o(history._getLabel().text()).equals("0/2")("Checking label text.")

                o(relays._getRows().length).equals(1)("Checking visible rows.")
                o(relays._getLabel().text()).equals("1/2")("Checking label text.")
                done()

            jQuery("#" + self.tabs._searchId).val("alpha")
            self.tabs.searchAll()

            self._redraw(f2)

        def relaysData():
            self.tabs.tabs[2].table._setData(self._Relays_Data)
            self._redraw(f1)

        self.tabs.tabs[0].table._setData(self._History_Data)
        self._redraw(relaysData)

    # ============================================== #

    def tableSort(self):
        """
        Tests table sorting.
        """
        a = {"id": "a", "blob": ""}
        a3 = {"id": "a", "blob": "3"}
        a1 = {"id": "a", "blob": "1"}
        b4 = {"id": "b", "blob": "4"}
        d2 = {"id": "d", "blob": "2"}
        c5 = {"id": "c", "blob": "5"}
        data = [a3, a, a1, b4, d2, c5]

        blobs = self.tabs.tabs[1]
        blobs.table._setData(data)

        field = blobs.table.fields[0]
        o(field.title).equals("DID")("Checking field header.")
        blobs.table.setSort(field)
        o(blobs.table._shownData).deepEquals([a3, a, a1, b4, c5, d2])("Checking row order.")

        blobs.table.setSort(field)
        o(blobs.table._shownData).deepEquals([d2, c5, b4, a1, a, a3])("Checking row order.")

        field = blobs.table.fields[1]
        o(field.title).equals("Blob")("Checking field header.")
        blobs.table.setSort(field)
        o(blobs.table._shownData).deepEquals([a, a1, d2, a3, b4, c5])("Checking row order.")

        blobs.table.setSort(field)
        o(blobs.table._shownData).deepEquals([c5, b4, a3, d2, a1, a])("Checking row order.")

    # ============================================== #

    def asyncSelectRows(self, done, timeout):
        """
        Tests row selection.

            Parameters:
            done - Automatically included promise done function
            timeout - Automatically included timeout
        """
        timeout(200)

        def f1():
            def f2():
                tab = self.tabs.currentTab()
                expected = tab.table._stringify(self._History_Data[0])
                actual = jQuery("#" + tab._detailsId).text()
                o(actual).equals(expected)("Checking selected row.")
                o(jQuery("#" + tab._copiedId).text()).equals("")("Checking copied data.")

                def f3():
                    expected = tab.table._stringify(self._History_Data[1])
                    actual = jQuery("#" + tab._detailsId).text()
                    o(actual).equals(expected)("Checking selected row.")
                    o(jQuery("#" + tab._copiedId).text()).equals("")("Checking copied data.")
                    done()

                self._clickRow(1, f3)

            self._clickRow(0, f2)

        self._setData(f1)

    # ============================================== #

    def asyncDetailsCopy(self, done, timeout):
        """
        Tests detail copy.

            Parameters:
            done - Automatically included promise done function
            timeout - Automatically included timeout
        """
        timeout(400)

        def f1():
            def f2():
                tab = self.tabs.currentTab()

                def f3():
                    expected = tab.table._stringify(self._History_Data[0])
                    o(jQuery("#" + tab._detailsId).text()).equals(expected)("Checking selected row.")
                    o(jQuery("#" + tab._copiedId).text()).equals(expected)("Checking copied data.")

                    def f4():
                        o(jQuery("#" + tab._detailsId).text()).equals(expected)("Checking selected row.")
                        o(jQuery("#" + tab._copiedId).text()).equals("")("Checking copied data.")
                        done()

                    self._clickId(tab._clearButtonId, f4)

                self._clickId(tab._copyButtonId, f3)

            self._clickRow(0, f2)

        self._setData(f1)

    # ============================================== #

    def asyncRowLimit(self, done):
        """
        Tests limiting rows.

            Parameters:
            done - Automatically included promise done function
        """
        table = self.tabs.currentTab().table

        def f1():
            history = self.tabs.tabs[0]
            rows = history._getRows()
            o(rows.length).equals(table.max_size + 1)("Checking number of rows.")
            o(rows.last().find("td").text()).equals(table._limitText())("Check last row.")
            o(history._getLabel().text()).equals("{0}/{1}".format(table.max_size, table.total))("Check label format.")
            done()

        table.max_size = 50
        temp = table._makeDummyData(table.max_size * 2)

        data = []
        for entry in temp:
            new = {"history":
                    {
                       "id": entry["did"],
                        "changed": entry["changed"],
                        "signer": entry["signer"],
                        "signers": entry["signers"]
                    },
                   "signatures": entry["signatures"]
                  }
            data.append(new)

        table._setData(data)
        self._redraw(f1)

# ================================================== #
#                        EOF                         #
# ================================================== #


import falcon
try:
    import simplejson as json
except ImportError:
    import json

from didery.help import helping


# Test data
tempDB = []
# tempDB = [
#     {
#         "title": "Invalid Signature.",
#         "msg": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= had an invalid rotation signature.",
#         "time": "2000-01-01T00:00:00+00:00"
#     },
#     {
#         "title": "Relay Unreachable.",
#         "msg": "Could not establish a connection with relay servers.",
#         "time": "2000-01-01T11:00:00+00:00"
#     }
# ]


class Error:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store

    """
    For manual testing of the endpoint:
        http localhost:8000/errors
    """
    @falcon.before(helping.parseQString)
    def on_get(self, req, resp):
        """
        Handle and respond to incoming GET request.
        :param req: Request object
        :param resp: Response object
        """
        offset = req.offset
        limit = req.limit

        if offset >= len(tempDB):
            resp.body = json.dumps({"data": []}, ensure_ascii=False)
            return

        body = {
            "data": tempDB[offset:offset + limit]
        }
        resp.append_header('X-Total-Count', len(tempDB))

        resp.body = json.dumps(body, ensure_ascii=False)



"""
core package

flo behaviors

"""
from __future__ import generator_stop

import importlib

_modules = ['resting', ]

for m in _modules:
    importlib.import_module(".{0}".format(m), package='didery.core')



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -m dideryd` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``didery.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``didery.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import os
import click
import ioflo.app.run

from ioflo.aid import odict
from ioflo.aid.consoling import VERBIAGE_NAMES


@click.command()
@click.option(
    '--port',
    '-p',
    multiple=False,
    default=8080,
    type=click.IntRange(1, 65535),
    help='port number the server should listen on'
)
@click.option(
    '--verbose',
    '-v',
    type=click.Choice(VERBIAGE_NAMES),
    default=VERBIAGE_NAMES[2],
    help='verbosity level'
)
@click.option(
    '--path',
    multiple=False,
    type=click.Path(file_okay=False, resolve_path=True, writable=True),
    help='path to the database folder'
)
def main(port, verbose, path):
    projectDirpath = os.path.dirname(
        os.path.dirname(
            os.path.abspath(
                os.path.expanduser(__file__)
            )
        )
    )
    floScriptpath = os.path.join(projectDirpath, "didery/flo/main.flo")

    click.echo("MAIN")
    """ Main entry point for ioserve CLI"""

    verbose = VERBIAGE_NAMES.index(verbose)

    ioflo.app.run.run(name="skedder",
                      period=0.125,
                      real=True,
                      retro=True,
                      filepath=floScriptpath,
                      behaviors=['didery.core'],
                      mode='',
                      username='',
                      password='',
                      verbose=verbose,
                      consolepath='',
                      statistics=False,
                      preloads=[
                          ('.main.server.port', odict(value=port)),
                          ('.main.server.db', odict(value=path))
                      ])



# ================================================== #
#                    TEST SERVER                     #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/21/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

__pragma__ ('alias', 'Global', 'global')


import didery.static.transcrypt.server as server
from tester import test

o = require("mithril/ospec/ospec")
sinon = require("sinon")

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

@test
class Server:
    """
    Class for testing server.
    """
    def beforeEach(self):
        """
        Sets up fake server and window.
        """
        self.testServer = sinon.createFakeServer()
        self.manager = server.Manager()
        window.XMLHttpRequest = Global.XMLHttpRequest

    # ============================================== #

    def afterEach(self):
        """
        Resets fake server.
        """
        self.testServer.restore()

    # ============================================== #

    def _respond(self, request, response):
        """
        Gives response to request.

            Parameters:
            request - HTTP request
            response - HTTP response
        """
        request.respond(200, {"Content-Type": "application/json"}, JSON.stringify(response))

    # ============================================== #

    def _respondTo(self, endpoint, data):
        """
        Gives response for specific endpoint.

            Parameters:
            endpoint - URL of API endpoint
            data - Data to be returned
        """
        self.testServer.respondWith(endpoint, lambda request: self._respond(request, data))

    # ============================================== #

    def basicRequest(self):
        """
        Tests basic server request.
        """
        endpoint = "/foo"

        def verify(request):
            o(request.method).equals("GET")("Checking basic request method.")
            o(request.url).equals(endpoint)("Checking basic request url.")
            request.respond(200)
        self.testServer.respondWith(verify)

        server.request(endpoint)
        self.testServer.respond()

    # ============================================== #

    def queryRequest(self):
        """
        Tests server request with query.
        """
        endpoint = "/foo"
        full_path = "/foo?one=1&two=2"

        def verify(request):
            o(request.method).equals("GET")("Checking query request method.")
            o(request.url).equals(full_path)("Checking basic request url.")
            request.respond(200)
        self.testServer.respondWith(verify)

        server.request(endpoint, one=1, two=2)
        self.testServer.respond()

    # ============================================== #

    def asyncErrors(self, done):
        """
        Tests errors request.

            Parameters:
            done - Automatically included promise done function
        """
        o(len(self.manager.errors.errors)).equals(0)("Checking initial errors list.")

        self._respondTo("/errors", {
            "data": [{
                    "title": "Invalid Signature.",
                    "msg": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= had an invalid rotation signature.",
                    "time": "2000-01-01T00:00:00+00:00"
                },
                {
                    "title": "Relay Unreachable.",
                    "msg": "Could not establish a connection with relay servers.",
                    "time": "2000-01-01T11:00:00+00:00"
                }]
            })

        def f1():
            o(len(self.manager.errors.errors)).equals(2)("Checking errors list after promise.")
            o(self.manager.errors.errors[0]).deepEquals({
                    "title": "Invalid Signature.",
                    "msg": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= had an invalid rotation signature.",
                    "time": "2000-01-01T00:00:00+00:00"
                })("Checking first error entry.")
            o(self.manager.errors.errors[1]).deepEquals({
                    "title": "Relay Unreachable.",
                    "msg": "Could not establish a connection with relay servers.",
                    "time": "2000-01-01T11:00:00+00:00"
                })("Checking second error entry.")
            done()

        self.testServer.autoRespond = True
        self.manager.errors.refreshErrors().then(f1)

    # ============================================== #

    def asyncHistory(self, done):
        """
        Tests history request.

            Parameters:
            done - Automatically included promise done function
        """
        o(len(self.manager.history.history)).equals(0)("Checking initial history list.")

        self._respondTo("/history", {
            "data": [{
                "history":
                {
                    "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                    "changed" : "2000-01-01T00:00:00+00:00",
                    "signer": 2,
                    "signers":
                    [
                        "Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA="
                    ]
                },
                "signatures":
                [
                    "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCg==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            }, {
                "history":
                {
                    "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                    "changed" : "2000-01-01T00:00:00+00:00",
                    "signer": 1,
                    "signers":
                    [
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY="
                    ]
                },
                "signatures":
                [
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            }]
        })

        def f1():
            o(len(self.manager.history.history)).equals(2)("Checking history list after promise.")
            o(self.manager.history.history[0]).deepEquals({
                "history":
                {
                    "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                    "changed" : "2000-01-01T00:00:00+00:00",
                    "signer": 2,
                    "signers":
                    [
                        "Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA="
                    ]
                },
                "signatures":
                [
                    "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCg==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            })("Checking first history entry.")
            o(self.manager.history.history[1]).deepEquals({
                "history":
                {
                    "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                    "changed" : "2000-01-01T00:00:00+00:00",
                    "signer": 1,
                    "signers":
                    [
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                        "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY="
                    ]
                },
                "signatures":
                [
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw==",
                    "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                ]
            })("Checking second history entry.")
            done()

        self.testServer.autoRespond = True
        self.manager.history.refreshHistory().then(f1)

    # ============================================== #

    def asyncOTPBlobs(self, done):
        """
        Tests blobs request.

            Parameters:
            done - Automatically included promise done function
        """
        o(len(self.manager.otpBlobs.blobs)).equals(0)("Checking initial blobs list.")

        self._respondTo("/blob", {
            "data": [{
                "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            },
            {
                "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            }]
        })

        def f1():
            o(len(self.manager.otpBlobs.blobs)).equals(2)("Checking blobs list after promise.")
            o(self.manager.otpBlobs.blobs[0]).deepEquals({
                "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            })("Checking first blob entry.")
            o(self.manager.otpBlobs.blobs[1]).deepEquals({
                "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
            })("Checking second blob entry.")
            done()

        self.testServer.autoRespond = True
        self.manager.otpBlobs.refreshBlobs().then(f1)

    # ============================================== #

    def asyncRelays(self, done):
        """
        Tests relays request.

            Parameters:
            done - Automatically included promise done function
        """
        o(len(self.manager.relays.relays)).equals(0)("Checking initial relay list.")

        self._respondTo("/relay", {
            "1": {
                "host address": "127.0.0.1",
                "port": 7541,
                "name": "alpha",
                "main": True,
                "uid": "1",
                "status": "connected"
            },
            "2": {
                "host address": "127.0.0.1",
                "port": 7542,
                "name": "beta",
                "main": False,
                "uid": "2",
                "status": "connected"
            }
        })

        def f1():
            o(len(self.manager.relays.relays)).equals(2)("Checking relays list after promise.")
            o(self.manager.relays.relays[0]).deepEquals({
                "host address": "127.0.0.1",
                "port": 7541,
                "name": "alpha",
                "main": True,
                "uid": "1",
                "status": "connected"
            })("Checking first relay entry.")
            o(self.manager.relays.relays[1]).deepEquals({
                "host address": "127.0.0.1",
                "port": 7542,
                "name": "beta",
                "main": False,
                "uid": "2",
                "status": "connected"
            })("Checking second relay entry.")
            done()

        self.testServer.autoRespond = True
        self.manager.relays.refreshRelays().then(f1)

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                      SERVER                        #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/03/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

# ================================================== #
#                 CONSTANTS & GLOBALS                #
# ================================================== #

DEFAULT_INTERVAL = 1000

# ================================================== #
#                     FUNCTIONS                      #
# ================================================== #

__pragma__("kwargs")
def request(path, **kwargs):
    """
    Performs a mithril GET request.

        Parameters:
        path - Endpoint string
        kwargs - Dictionary of path query arguments

        Returns:
        Promise from m.request
    """
    path += "?"
    for key, value in kwargs.items():
        path += key + "=" + str(value) + "&"
    path = path[:-1]
    return m.request(path)

# ================================================== #

def onlyOne(func, interval=1000):
    """
    Enforces the promise function. Never called more than once
    per interval.

        Parameters:
        func - Executor function
        interval - Promise interval in milliseconds
    """
    scope = {"promise": None, "lastCalled": 0}

    def wrap():
        now = __new__(Date())
        if scope.promise != None and now - scope.lastCalled < interval:
            return scope.promise

        scope.lastCalled = now

        def f(resolve, reject):
            p = func()
            p.then(resolve)
            p.catch(reject)
        scope.promise = __new__(Promise(f))
        return scope.promise
    return wrap
__pragma__("nokwargs")

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

def clearArray(a):
    """
    Clears an array/list.

        Parameters:
        a - Array/List to be cleared
    """
    while len(a):
        a.pop()

# ================================================== #

class Manager:
    """
    Class for managing server calls.
    """
    def __init__(self):
        """
        Initialize Manager object. Loads all server calls.
        """
        self.errors = Errors()
        self.history = History()
        self.otpBlobs = OTPBlobs()
        self.relays = Relays()

# ================================================== #

class Errors:
    """
    Class for error server call.
    """
    Refresh_Interval = DEFAULT_INTERVAL
    # Refresh rates should never exceed the default interval rate.

    def __init__(self):
        """
        Initialize Errors object.
        """
        self.errors = []
        self.refreshErrors = onlyOne(self._refreshErrors, interval=self.Refresh_Interval)

    # ============================================== #

    def _refreshErrors(self):
        """
        Clears error array and retrieves fresh error data.
        """

        clearArray(self.errors)
        return request("/errors").then(self._parseAll)

    # ============================================== #

    def _parseAll(self, data):
        """
        Parses returned error data.

            Parameters:
            data - Returned error data

            Example:
            {
                "data": [{
                        "title": "Invalid Signature.",
                        "msg": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= had an invalid rotation signature.",
                        "time": "2000-01-01T00:00:00+00:00"
                    },
                    {
                        "title": "Relay Unreachable.",
                        "msg": "Could not establish a connection with relay servers.",
                        "time": "2000-01-01T11:00:00+00:00"
                    }]
            }
        """

        for error in data['data']:
            self.errors.append(error)

# ================================================== #

class History:
    """
    Class for history server call.
    """
    Refresh_Interval = DEFAULT_INTERVAL
    # Refresh rates should never exceed the default interval rate.

    def __init__(self):
        """
        Initialize History object.
        """
        self.history = []
        self.refreshHistory = onlyOne(self._refreshHistory, interval=self.Refresh_Interval)

    # ============================================== #

    def _refreshHistory(self):
        """
        Clears history array and retrieves fresh history data.
        """

        clearArray(self.history)
        return request("/history").then(self._parseAll)

    # ============================================== #

    def _parseAll(self, data):
        """
        Parses returned history data.

            Parameters:
            data - Returned history data

            Example:
            {
                "data": [{
                    "history":
                    {
                        "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                        "changed" : "2000-01-01T00:00:00+00:00",
                        "signer": 2,
                        "signers":
                        [
                            "Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                            "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                            "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                            "3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA="
                        ]
                    },
                    "signatures":
                    [
                        "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCg==",
                        "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                    ]
                }, {
                    "history":
                    {
                        "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                        "changed" : "2000-01-01T00:00:00+00:00",
                        "signer": 1,
                        "signers":
                        [
                            "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                            "Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=",
                            "dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY="
                        ]
                    },
                    "signatures":
                    [
                        "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw==",
                        "o9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw=="
                    ]
                }]
            }

        """

        for history in data['data']:
            self.history.append(JSON.parse(history))

# ================================================== #

class OTPBlobs:
    """
    Class for otp blob server call.
    """
    Refresh_Interval = DEFAULT_INTERVAL

    # Refresh rates should never exceed the default interval rate.

    def __init__(self):
        """
        Initialize OTPBlob object.
        """
        self.blobs = []
        self.refreshBlobs = onlyOne(self._refreshBlobs, interval=self.Refresh_Interval)

    # ============================================== #

    def _refreshBlobs(self):
        """
        Clears blob array and retrieves fresh blob data.
        """

        clearArray(self.blobs)
        return request("/blob").then(self._parseAll)

    # ============================================== #

    def _parseAll(self, data):
        """
        Parses returned blob data.

            Parameters:
            data - Returned blob data

            Example:
            {
                "data": [{
                    "id": "did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=",
                    "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
                },
                {
                    "id": "did:igo:dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=",
                    "blob": "AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
                }]
            }
        """

        for blob in data['data']:
            self.blobs.append(JSON.parse(blob))

# ================================================== #

class Relays:
    """
    Class for relays server call.
    """
    Refresh_Interval = DEFAULT_INTERVAL
    # Refresh rates should never exceed the default interval rate.

    def __init__(self):
        """
        Initialize Relays object.
        """
        self.relays = []
        self.refreshRelays = onlyOne(self._refreshRelays, interval=self.Refresh_Interval)

    # ============================================== #

    def _refreshRelays(self):
        """
        Clears relay array and retrieves fresh relay data.
        """

        clearArray(self.relays)
        return request("/relay").then(self._parseAll)

    # ============================================== #

    def _parseAll(self, data):
        """
        Parses returned relay data.

            Parameters:
            data - Returned relay data

            Example:
            {
                "1": {
                    "host address": "127.0.0.1",
                    "port": 7541,
                    "name": "alpha",
                    "main": True,
                    "uid": "1",
                    "status": "connected"
                },
                "2": {
                    "host address": "127.0.0.1",
                    "port": 7542,
                    "name": "beta",
                    "main": False,
                    "uid": "2",
                    "status": "connected"
                }
            }
        """

        for relay in dict(data).items():
            self.relays.append(relay[1])

# ================================================== #
#                       MAIN                         #
# ================================================== #

manager = Manager()

# ================================================== #
#                        EOF                         #
# ================================================== #


import falcon
import arrow
import time
try:
    import simplejson as json
except ImportError:
    import json

from collections import OrderedDict
from ..help import helping
from .. import didering
from ..db import dbing as db


def basicValidation(req, resp, resource, params):
    raw = helping.parseReqBody(req)
    body = req.body

    required = ["id", "changed", "signer", "signers"]
    helping.validateRequiredFields(required, body)

    signature = req.get_header("Signature", required=True)
    sigs = helping.parseSignatureHeader(signature)
    req.signatures = sigs

    if len(sigs) == 0:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Empty Signature header.')

    try:
        if not isinstance(body['signers'], list):
            body['signers'] = json.loads(
                body['signers'].replace("'", '"')
            )
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'signers field must be a list or array.')

    if body['id'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'id field cannot be empty.')

    if body['changed'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'changed field cannot be empty.')

    try:
        req.body['signer'] = int(body['signer'])
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'signer field must be a number.')

    try:
        dt = arrow.get(body["changed"])
    except arrow.parser.ParserError as ex:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'ISO datetime could not be parsed.')

    for value in body['signers']:
        if value == "":
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'signers keys cannot be empty.')

    return raw, sigs


def validatePost(req, resp, resource, params):
    """
    Validate incoming POST request and prepare
    body of request for processing.
    :param req: Request object
    """
    # server crashes without this if someone adds anything after /history
    if params:
        raise falcon.HTTPError(falcon.HTTP_404)

    raw, sigs = basicValidation(req, resp, resource, params)
    body = req.body

    sig = sigs.get('signer')  # str not bytes
    if not sig:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing "signer" tag and signature.')

    try:
        didkey = helping.extractDidParts(body['id'])
    except ValueError as ex:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               "Invalid did format. {}".format(str(ex)))

    if len(body['signers']) < 2:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'signers field must contain at least the current public key and its first pre-rotation.')

    # If it is decided that revocation on inception is allowed we can remove this check
    for value in body['signers']:
        if value is None:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'signers keys cannot be null on inception.')

    if body['signer'] != 0:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'signer field must equal 0 on creation of new rotation history.')

    index = body['signer']
    try:
        helping.validateSignedResource(sig, raw, body['signers'][index])
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request body and signature. {}.'.format(ex))

    # Prevent bad actors from trying to commandeer a DID before its owner posts it
    if didkey != body['signers'][0]:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'The DIDs key must match the first key in the signers field.')


def validatePut(req, resp, resource, params):
    if 'did' not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'DID value missing from url.')

    raw, sigs = basicValidation(req, resp, resource, params)
    body = req.body

    signer = sigs.get('signer')  # str not bytes
    if not signer:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing signature for "signer".')

    rotation = sigs.get('rotation')  # str not bytes
    if not rotation:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing signature for "rotation".')

    if len(body['signers']) < 3:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'PUT endpoint is for rotation events. Must contain at least the original key, a '
                               'current signing key, and a pre-rotated key.')

    if body['signer'] < 1 or body['signer'] >= len(body['signers']):
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               '"signer" cannot reference the first or last key in the "signers" '
                               'field on PUT requests.')

    if body['signer'] + 1 == len(body['signers']) and body['signers'][body["signer"]] is not None:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'Missing pre rotated key in the signers field.')

    # Prevent did data from being clobbered
    if params['did'] != body['id']:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'Url did must match id field did.')

    for key, value in enumerate(body['signers']):
        if value is None and body['signer'] != key:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'signers keys cannot be null unless revoking a key.')

    if body["signers"][body["signer"]] is None:
        index = body['signer'] - 1
    else:
        index = body['signer']

    try:
        helping.validateSignedResource(rotation, raw, body['signers'][index])
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request signature for rotation field. {}.'.format(ex))

    try:
        helping.validateSignedResource(signer, raw, body['signers'][index-1])
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request signature for signer field. {}.'.format(ex))


def validateDelete(req, resp, resource, params):
    if 'did' not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'DID value missing from url.')

    raw = helping.parseReqBody(req)
    signature = req.get_header("Signature", required=True)
    sigs = helping.parseSignatureHeader(signature)
    req.signatures = sigs
    body = req.body

    if len(sigs) == 0:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Empty Signature header.')

    signer = sigs.get('signer')  # str not bytes
    if not signer:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing signature for "signer".')

    # Prevent did data from being clobbered
    if params['did'] != body['id']:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'Url did must match id field did.')

    history = db.getHistory(params['did'])
    req.history = history
    if history is None:
        raise falcon.HTTPError(falcon.HTTP_404)

    index = history['history']['signer']
    vk = history['history']['signers'][index]

    if vk is not None:
        try:
            helping.validateSignedResource(signer, raw, vk)
        except didering.ValidationError as ex:
            raise falcon.HTTPError(falcon.HTTP_401,
                                   'Authorization Error',
                                   'Could not validate the request signature for signer field. {}.'.format(ex))

    else:
        try:
            helping.validateSignedResource(signer, raw, history['history']['signers'][index - 1])
        except didering.ValidationError as ex:
            raise falcon.HTTPError(falcon.HTTP_401,
                                   'Authorization Error',
                                   'Could not validate the request signature for signer field. {}.'.format(ex))


class History:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store

    """
    For manual testing of the endpoint:
        http localhost:8000/history/did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=
        http localhost:8000/history
    """
    @falcon.before(helping.parseQString)
    def on_get(self, req, resp, did=None):
        """
        Handle and respond to incoming GET request.
        :param req: Request object
        :param resp: Response object
        :param did: string
            URL parameter specifying a rotation history
        """
        offset = req.offset
        limit = req.limit

        count = db.historyCount()

        if did is not None:
            body = db.getHistory(did)
            if body is None:
                raise falcon.HTTPError(falcon.HTTP_404)
        else:
            if offset >= count:
                resp.body = json.dumps({}, ensure_ascii=False)
                return

            body = db.getAllHistories(offset, limit)

            resp.append_header('X-Total-Count', count)

        resp.body = json.dumps(body, ensure_ascii=False)


    """
    For manual testing of the endpoint:
        http POST localhost:8000/history id="did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=" changed="2000-01-01T00:00:00+00:00" signer=2 signers="['Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=', 'Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=', 'dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=', '3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA=']"
    """
    @falcon.before(validatePost)
    def on_post(self, req, resp):
        """
        Handle and respond to incoming POST request.
        :param req: Request object
        :param resp: Response object
        """
        result_json = req.body
        sigs = req.signatures
        did = result_json['id']

        if db.getHistory(did) is not None:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Resource Already Exists',
                                   'Resource with did "{}" already exists. Use PUT request.'.format(result_json['id']))

        # TODO: review signature validation for any holes
        response_json = db.saveHistory(did, result_json, sigs)
        db.saveEvent(did, result_json, sigs)

        resp.body = json.dumps(response_json, ensure_ascii=False)
        resp.status = falcon.HTTP_201

    """
    For manual testing of the endpoint:
        http PUT localhost:8000/history/did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= id="did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=" changed="2000-01-01T00:00:00+00:00" signer=2 signers="['Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=', 'Xq5YqaL6L48pf0fu7IUhL0JRaU2_RxFP0AL43wYn148=', 'dZ74MLZXD-1QHoa73w9pQ9GroAvxqFi2RTZWlkC0raY=', '3syVH2woCpOvPF0SD9Z0bu_OxNe2ZgxKjTQ961LlMnA=']"
    """
    @falcon.before(validatePut)
    def on_put(self, req, resp, did):
        """
            Handle and respond to incoming PUT request.
            :param req: Request object
            :param resp: Response object
            :param did: decentralized identifier
        """
        result_json = req.body
        sigs = req.signatures

        resource = db.getHistory(did)

        if resource is None:
            raise falcon.HTTPError(falcon.HTTP_404)

        # make sure time in changed field is greater than existing changed field
        last_changed = arrow.get(resource['history']['changed'])
        new_change = arrow.get(result_json['changed'])

        if last_changed >= new_change:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   '"changed" field not later than previous update.')

        # validate that previously rotated keys are not changed with this request
        current = resource['history']['signers']
        update = result_json['signers']

        if len(update) <= len(current):
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'signers field is missing keys.')

        for key, val in enumerate(current):
            if update[key] != val:
                raise falcon.HTTPError(falcon.HTTP_400,
                                       'Validation Error',
                                       'signers field missing previously verified keys.')

        # without these checks a hacker can skip past the validated signatures and insert their own keys
        if resource['history']['signer'] + 1 != result_json['signer']:
            if result_json['signers'][result_json['signer']] is not None:
                raise falcon.HTTPError(falcon.HTTP_400,
                                       'Validation Error',
                                       'signer field must be one greater than previous.')

        # TODO: review signature validation for any holes
        response_json = db.saveHistory(did, result_json, sigs)
        db.saveEvent(did, result_json, sigs)

        resp.body = json.dumps(response_json, ensure_ascii=False)

    @falcon.before(validateDelete)
    def on_delete(self, req, resp, did):
        """
            Handle and respond to incoming PUT request.
            :param req: Request object
            :param resp: Response object
            :param did: decentralized identifier
        """
        resource = req.history

        success = db.deleteHistory(did)
        db.deleteEvent(did)

        if not success:
            raise falcon.HTTPError(falcon.HTTP_500,
                                   'Deletion Error',
                                   'Error while attempting to delete the resource.')

        resp.body = json.dumps({"deleted": resource}, ensure_ascii=False)


class HistoryStream:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store
        self.id = 0
        self.history = None

    def historyGenerator(self, did=None):
        if did is None:
            count = db.historyCount()
            if self.history == count:
                time.sleep(2)
            else:
                self.history = count
                history = db.getAllHistories(limit=self.history)
                temp = []
                for entry in history['data']:
                    temp.append(json.loads(entry.decode('utf-8')))

                history = {"data": temp}
                yield bytes("\nid:" + str(self.id) + "\nevent:message\ndata:" + str(history) + "\n\n", "ascii")
                self.id += 1
                time.sleep(2)

        else:
            value = db.getHistory(did)
            if value is None:
                raise falcon.HTTPError(falcon.HTTP_404)

            if self.history == value:
                time.sleep(2)
            else:
                self.history = value
                print(self.history)
                yield bytes("\nid:" + str(self.id) + "\nevent:message\ndata:" + str(self.history) + "\n\n", "ascii")
                self.id += 1
                time.sleep(2)

    def on_get(self, req, resp, did=None):
        resp.status = falcon.HTTP_200
        resp.content_type = "text/event-stream"
        resp.stream = self.historyGenerator(did)



import falcon
import arrow
import time
try:
    import simplejson as json
except ImportError:
    import json

from collections import OrderedDict
from ..help import helping
from .. import didering
from ..db import dbing as db

class Event:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store

    #@falcon.before(helping.parseQString)
    def on_get(self, req, resp, did=None):
        """
        Handle and respond to incoming GET request.
        :param req: Request object
        :param resp: Response object
        :param did: string
            URL parameter specifying a rotation history
        """
        #offset = req.offset
        #limit = req.limit

        count = db.eventCount()

        if did is not None:
            body = db.getEvent(did)
            if body is None:
                raise falcon.HTTPError(falcon.HTTP_404)
        else:
            #if offset >= count:
                #resp.body = json.dumps({}, ensure_ascii=False)
                #return


            body = db.getAllEvents(limit=count)

            resp.append_header('X-Total-Count', count)

        resp.body = json.dumps(body, ensure_ascii=False)


from didery.controllers import history as histories
from didery.controllers import otp_blob as blobs
from didery.controllers import relays
from didery.controllers import errors
from didery.controllers import static
from didery.controllers import events

STATIC_BASE_PATH = "/static"
DEFAULT_STATIC_BASE_PATH = "/"
STREAM_BASE_PATH = "/stream"
HISTORY_BASE_PATH = "/history"
BLOB_BASE_PATH = "/blob"
RELAY_BASE_PATH = "/relay"
ERRORS_BASE_PATH = "/errors"
EVENTS_BASE_PATH = "/event"


class CORSMiddleware:
    def process_request(self, req, resp):
        resp.set_header('Access-Control-Max-Age:', '3600')
        resp.set_header('Access-Control-Allow-Origin', '*')
        resp.set_header('Access-Control-Allow-Methods',
                                   'PUT, GET, POST, DELETE, HEAD, OPTIONS')
        resp.set_header('Access-Control-Allow-Headers',
                                   'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token, X-Auth-Token, Signature')


def loadEndPoints(app, store):
    """
    Add Rest endpoints to a falcon.API object by mapping the API's routes.
    :param app: falcon.API object
    :param store: Store
        ioflo datastore
    """

    sink = static.StaticSink()
    app.add_sink(sink, prefix=DEFAULT_STATIC_BASE_PATH)

    history = histories.History(store)
    app.add_route('{}/{{did}}'.format(HISTORY_BASE_PATH), history)
    app.add_route('{}'.format(HISTORY_BASE_PATH), history)

    historyStream = histories.HistoryStream(store)
    app.add_route('{}{}/{{did}}'.format(STREAM_BASE_PATH, HISTORY_BASE_PATH), historyStream)
    app.add_route('{}{}'.format(STREAM_BASE_PATH, HISTORY_BASE_PATH), historyStream)

    blob = blobs.OtpBlob(store)
    app.add_route('{}/{{did}}'.format(BLOB_BASE_PATH), blob)
    app.add_route('{}'.format(BLOB_BASE_PATH), blob)

    relay = relays.Relay(store)
    app.add_route('{}/{{uid}}'.format(RELAY_BASE_PATH), relay)
    app.add_route('{}'.format(RELAY_BASE_PATH), relay)

    error = errors.Error(store)
    app.add_route('{}'.format(ERRORS_BASE_PATH), error)

    event = events.Event(store)
    app.add_route('{}/{{did}}'.format(EVENTS_BASE_PATH), event)
    app.add_route('{}'.format(EVENTS_BASE_PATH), event)



# ================================================== #
#                     DASHBOARD                      #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 04/26/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

import didery.static.transcrypt.components.tabs as tabs
import didery.static.transcrypt.components.searcher as searcher

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class Manager:
    """
    Class for managing all displayed tabs.
    """

    def __init__(self):
        """
        Initialize Tabs object. Load in all tabs and setup
        document level functions.
        """
        self.tabs = [tabs.History(), tabs.Blobs(), tabs.Relays(), tabs.Errors()]
        self._searchId = "search-input"
        self.searcher = searcher.Searcher()
        self._refreshing = False
        self._refreshPromise = None

        jQuery(document).ready(lambda: jQuery('.menu > a.item').tab())

        self.refresh()

    # ============================================== #

    def refresh(self):
        """
        Retrieves server data and populates our tabs and tables.
        """
        if self._refreshing:
            return self._refreshPromise
        self._refreshing = True

        promises = []
        for tab in self.tabs:
            promises.append(tab.table.refresh())

        def done():
            self._refreshing = False

        self._refreshPromise = Promise.all(promises)
        self._refreshPromise.then(done)
        self._refreshPromise.catch(done)
        return self._refreshPromise

    # ============================================== #

    def currentTab(self):
        """
        Returns current tab.
        """
        active = jQuery(".menu a.item.active")
        data_tab = active.attr("data-tab")
        for tab in self.tabs:
            if tab.DataTab == data_tab:
                return tab
        return None

    # ============================================== #

    def searchAll(self):
        """
        Initiates searching across all tabs based on current search string.
        """
        text = jQuery("#" + self._searchId).val()
        self.searcher.setSearch(text)

        for tab in self.tabs:
            tab.table.setFilter(self.searcher.search)

    # ============================================== #

    def view(self):
        """
        Returns markup for view.
        """
        menu_items = []
        tab_items = []
        for tab in self.tabs:
            menu_items.append(tab.menu_item())
            tab_items.append(tab.tab_item())

        return m("div",
                 m("div.ui.top.attached.tabular.menu",
                   m("a.item.tab.hide",
                     m("span.menu-item-text", "Server Status"),
                     m("i.chart.bar.icon"),
                     m("div.ui.label.small.menu-item-number", "0/0")),
                   menu_items,
                   m("div.right.menu",
                     m("div.item",
                       m("form", {"onsubmit": self.searchAll},
                         m("div#search.ui.transparent.icon.input",
                           m("input[type=text][placeholder=Search...]", {"id": self._searchId}),
                           m("button.ui.icon.button[type=submit]",
                             m("i.search.link.icon"))))))),
                 tab_items)

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                       FIELDS                       #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/08/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class Field:
    """
    Class for table field/columns.
    """
    Title = None
    Length = 4

    __pragma__("kwargs")
    def __init__(self, title=None, length=None):
        """
        Initialize Field object. Set title and length.
        """
        self.title = self.Title
        if title is not None:
            self.title = title

        self.mlength = self.Length
        if length is not None:
            self.mlength = length

        self.name = self.title.lower()

    __pragma__("nokwargs")

    # ============================================== #

    def format(self, data):
        """
        Formats the data to a string matching the expected view for this field.

            Parameters:
            data - Data object to be formatted
        """
        return str(data)

    # ============================================== #

    def shorten(self, string):
        """
        Shortens the string to an appropriate length for display.

            Parameters:
            string - String to be shortened
        """
        return string

    # ============================================== #

    def view(self, data):
        """
        Returns a vnode <td> suitable for display in a table.

            Parameters:
            data - Data object
        """
        if data == None:
            data = ""
        formatted = self.format(data)
        return m("td", {"title": formatted}, self.shorten(formatted))

# ================================================== #

class FillField(Field):
    """
    Field that should "use remaining space" for display.
    """
    Length = 100

    def view(self, data):
        node = super().view(data)
        node.attrs["class"] = "fill-space"
        return node

# ================================================== #

class DateField(Field):
    """
    Field for displaying dates.
    """
    Length = 12
    Title = "Date"

# ================================================== #

class EpochField(DateField):
    """
    Field for displaying time since the epoch.
    """
    def format(self, data):
        # Make format match that of other typical dates from server
        data = __new__(Date(data / 1000)).toISOString()
        return super().format(data)

# ================================================== #

class IDField(Field):
    """
    Field for displaying ids.
    """
    Length = 4
    Title = "UID"
    Header = ""

    def format(self, string):
        if string.startswith(self.Header):
            string = string[len(self.Header):]

        return super().format(string)

# ================================================== #

class DIDField(IDField):
    """
    Field for displaying dids.
    """
    Header = "did:dad:"
    Title = "DID"

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                       TESTER                       #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/08/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

__pragma__ ('alias', 'Global', 'global')

jsdom = require('jsdom')
Global.window = __new__(jsdom.JSDOM()).window
Global.document = window.document
Global.XMLHttpRequest = window.XMLHttpRequest
Global.jQuery = require("jquery")
Global.m = require("mithril")
o = require("mithril/ospec/ospec")

# ================================================== #
#                     FUNCTIONS                      #
# ================================================== #

def test(Cls):
    """
    Runs tests.
    """
    class Wrapper:
        """
        Helper decorator for making test writing easier. Wrap around a
        class to treat the class as though o.spec were called on it. Treats
        any public methods as tests to be run through o(name, method).
        Methods or nested classes starting with an upper case are assumed
        to be nested test groups, and have test() called on them
        automatically. Special o methods (e.g. "beforeEach") are translated
        to the appropriate o call.
        """
        def __init__(self):
            original = Cls()

            def dospec():
                funcs = []

                __pragma__('jsiter')
                for key in original:
                    if key.startswith("_"):
                        continue

                    obj = original[key]

                    if key[0].isupper():
                        test(obj)
                        continue

                    if key.startswith("async"):
                        def makeScope(fun):
                            return lambda done, timeout: fun(done, timeout)
                        obj = makeScope(obj)
                        key = key[len("async"):]

                    if key in ["beforeEach", "BeforeEach"]:
                        o.beforeEach(obj)
                    elif key in ["before", "Before"]:
                        o.before(obj)
                    elif key in ["afterEach", "AfterEach"]:
                        o.afterEach(obj)
                    elif key in ["after", "After"]:
                        o.after(obj)
                    else:
                        funcs.append((key, obj))
                __pragma__('nojsiter')

                for name, func in funcs:
                    o(name, func)

            o.spec(Cls.__name__, dospec)
    Wrapper()
    return Wrapper

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                     DASHBOARD                      #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/04/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

from __future__ import generator_stop
import falcon
import mimetypes
import os

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class StaticSink(object):
    """
    Class for Falcon sink endpoint serving static files.
    """
    def __init__(self, *pa, **kwa):
        super().__init__(*pa, **kwa)
        self.projectDirpath = os.path.dirname(
                os.path.dirname(
                    os.path.abspath(
                        os.path.expanduser(__file__))))
        self.staticDirpath = os.path.join(self.projectDirpath, "static")

    def __call__(self, req, rep):
        path = req.path
        splits = path.split("/")[1:]
        if not splits[0]:
            splits = splits[1:]
        if splits and splits[0] == "static":
            splits = splits[1:]
        if not splits:
            filepath = "main.html"
        else:
            filepath = "/".join(splits)
        filepath = os.path.join(self.staticDirpath, filepath)
        if not os.path.exists(filepath):
            raise falcon.HTTPError(falcon.HTTP_NOT_FOUND,
                            'Missing Resource',
                            'File "{}" not found or forbidden'.format(filepath))
        filetype = mimetypes.guess_type(filepath, strict=True)[0]
        rep.set_header("Content-Type", "{}; charset=UTF-8".format(filetype))
        rep.status = falcon.HTTP_200
        with open(filepath, 'rb') as f:
            rep.body = f.read()

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                        TABS                        #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/08/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

import didery.static.transcrypt.components.tabledtab as tabledtab
import didery.static.transcrypt.components.tables as tables

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class Errors(tabledtab.TabledTab):
    """
    Class for errors tab.
    """
    Name = "Errors"
    Icon = "i.exclamation.circle.icon"
    DataTab = "errors"
    Active = False

    def setup_table(self):
        """
        Sets up errors table.
        """
        self.table = tables.ErrorsTable()

# ================================================== #

class Relays(tabledtab.TabledTab):
    """
    Class for relays tab.
    """
    Name = "Relays"
    Icon = "i.server.icon"
    DataTab = "relays"
    Active = False

    def setup_table(self):
        """
        Sets up relays table.
        """
        self.table = tables.RelaysTable()

# ================================================== #

class Blobs(tabledtab.TabledTab):
    """
   Class for encrypted blobs tab.
    """
    Name = "Encrypted Blobs"
    Icon = "i.unlock.alternate.icon"
    DataTab = "blobs"
    Active = False

    def setup_table(self):
        """
        Sets up blob table.
        """
        self.table = tables.BlobsTable()

# ================================================== #

class History(tabledtab.TabledTab):
    """
   Class for public keys tab.
    """
    Name = "Public Keys"
    Icon = "i.key.icon"
    DataTab = "history"
    Active = True

    def setup_table(self):
        """
        Sets up history table.
        """
        self.table = tables.HistoryTable()

# ================================================== #
#                        EOF                         #
# ================================================== #


import falcon
import uuid
import itertools
import arrow

try:
    import simplejson as json
except ImportError:
    import json

from ..help import helping


tempDB = {}


def basicValidation(req, resp, resource, params):
    """
    Validate incoming POST request and prepare
    body of request for processing.
    :param req: Request object
    """

    helping.parseReqBody(req)
    body = req.body

    required = ["host_address", "port", "name", "changed"]
    helping.validateRequiredFields(required, body)

    if body['host_address'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'host_address field cannot be empty.')

    if body['port'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'port field cannot be empty.')

    if body['name'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'name field cannot be empty.')

    if body['changed'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'changed field cannot be empty.')

    if 'main' in body:
        if body['main'] == "":
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'main field cannot be empty.')

        if type(body['main']) != bool:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   'main field must be a boolean value.')

    try:
        body['port'] = int(body['port'])
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'port field must be a number.')

    if body['port'] < 1 or body['port'] > 65535:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'port field must be a number between 1 and 65535.')


def validatePost(req, resp, resource, params):
    basicValidation(req, resp, resource, params)

    if 'uid' in req.body:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'If uid is known use PUT.')


def validatePut(req, resp, resource, params):
    basicValidation(req, resp, resource, params)

    if "uid" not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'uid required in url.')

    if "uid" not in req.body or req.body['uid'] == "":
        req.body['uid'] = params['uid']

    if params['uid'] != req.body['uid']:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'uid in url must match uid in body.')


def validateDelete(req, resp, resource, params):
    if "uid" not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'uid required in url.')


class Relay:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store

    """
    For manual testing of the endpoint:
        http localhost:8000/relay
    """
    @falcon.before(helping.parseQString)
    def on_get(self, req, resp):
        """
        Handle and respond to incoming GET request.
        :param req: Request object
        :param resp: Response object
        """
        offset = req.offset
        limit = req.limit

        if offset >= len(tempDB):
            resp.body = json.dumps({}, ensure_ascii=False)
            return

        body = dict(itertools.islice(tempDB.items(), offset, offset+limit))

        resp.append_header('X-Total-Count', len(tempDB))
        resp.body = json.dumps(body, ensure_ascii=False)

    """
    For manual testing of the endpoint:
        http POST localhost:8000/relay host_address="127.0.0.1" port=7541 name="alpha" main=true
    """
    @falcon.before(validatePost)
    def on_post(self, req, resp):
        """
        Handle and respond to incoming POST request.
        :param req: Request object
        :param resp: Response object
        """
        uid = uuid.uuid4().hex
        result_json = req.body

        if 'main' not in result_json:
            result_json['main'] = False

        result_json["uid"] = uid
        result_json["status"] = "connected"

        tempDB[uid] = result_json

        resp.body = json.dumps(result_json, ensure_ascii=False)
        resp.status = falcon.HTTP_201

    """
    For manual testing of the endpoint:
        http PUT localhost:8000/relay/1 host_address="127.0.0.1" port=7541 name="alpha" main=true
    """
    @falcon.before(validatePut)
    def on_put(self, req, resp, uid):
        """
        Handle and respond to incoming PUT request.
        :param req: Request object
        :param resp: Response object
        :param uid: servers unique id
        """
        result_json = req.body

        if uid not in tempDB:
            raise falcon.HTTPError(falcon.HTTP_404)

        resource = tempDB[uid]

        cdt = arrow.get(resource['changed'])
        udt = arrow.get(result_json['changed'])

        if cdt >= udt:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   '"changed" field not later than previous update.')

        resource['host_address'] = result_json['host_address']
        resource['port'] = result_json['port']
        resource['name'] = result_json['name']
        resource['changed'] = result_json['changed']

        if 'main' in result_json:
            resource['main'] = result_json['main']

        tempDB[uid] = resource

        resp.body = json.dumps(resource, ensure_ascii=False)

    """
    For manual testing of the endpoint:
        http DELETE localhost:8000/relay/10
    """
    @falcon.before(validateDelete)
    def on_delete(self, req, resp, uid):
        """
        Handle and respond to incoming DELETE request.
        :param req: Request object
        :param resp: Response object
        :param uid: servers unique id
        """
        if uid not in tempDB:
            raise falcon.HTTPError(falcon.HTTP_404)

        del tempDB[uid]



import ioflo.app.run


def main():
    print("MAIN")
    """ Main entry point for ioserve CLI"""
    # from didery import __version__
    args = ioflo.app.run.parseArgs(version="0.0.1")  # inject  version here

    ioflo.app.run.run(  name=args.name,
                        period=float(args.period),
                        real=args.realtime,
                        retro=args.retrograde,
                        filepath=args.filename,
                        behaviors=args.behaviors,
                        mode=args.parsemode,
                        username=args.username,
                        password=args.password,
                        verbose=args.verbose,
                        consolepath=args.console,
                        statistics=args.statistics)

if __name__ == '__main__':
    main()



import lmdb
import os
try:
    import simplejson as json
except ImportError:
    import json


MAX_DB_COUNT = 8

DATABASE_DIR_PATH = "/var/didery/db"
ALT_DATABASE_DIR_PATH = os.path.join('~', '.consensys/didery/db')

DB_EVENT_HISTORY_NAME = b'event_history'
DB_KEY_HISTORY_NAME = b'key_history'
DB_OTP_BLOB_NAME = b'otp_blob'

gDbDirPath = None   # database directory location has not been set up yet
dideryDB = None    # database environment has not been set up yet


def setupDbEnv(baseDirPath=None, port=8080):
    """
    Setup the module globals gDbDirPath, and dideryDB using baseDirPath
    if provided otherwise use DATABASE_DIR_PATH
    :param port: int
        used to differentiate dbs for multiple didery servers running on the same computer
    :param baseDirPath: string
        directory where the database is located
    """
    global gDbDirPath, dideryDB

    if not baseDirPath:
        baseDirPath = "{}{}".format(DATABASE_DIR_PATH, port)

    baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
    if not os.path.exists(baseDirPath):
        try:
            os.makedirs(baseDirPath)
        except OSError as ex:
            baseDirPath = "{}{}".format(ALT_DATABASE_DIR_PATH, port)
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)
    else:
        if not os.access(baseDirPath, os.R_OK | os.W_OK):
            baseDirPath = "{}{}".format(ALT_DATABASE_DIR_PATH, port)
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)

    gDbDirPath = baseDirPath  # set global

    dideryDB = lmdb.open(gDbDirPath, max_dbs=MAX_DB_COUNT)
    dideryDB.open_db(DB_EVENT_HISTORY_NAME)
    dideryDB.open_db(DB_KEY_HISTORY_NAME)
    dideryDB.open_db(DB_OTP_BLOB_NAME)

    return dideryDB

def eventCount():
    """
        Gets a count of the number of entries in the table

        :return: int count
    """
    subDb = dideryDB.open_db(DB_EVENT_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        return txn.stat(subDb)['entries']

def getEvent(did):
    """
        Find and return an event history matching the supplied did.

        :param did: string
            W3C did identifier for history object
        :return: dict
    """
    subDb = dideryDB.open_db(DB_EVENT_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        raw_data = txn.get(did.encode())

        if raw_data is None:
            return None

        return json.loads(raw_data)


def saveEvent(did, data, sigs):
    """
        Store an event and signatures

        :param did: string
            W3C did string
        :param data: dict
            A dict containing the rotation history and signatures

    """
    db_entry = []
    certifiable_data = {
        "event": data,
        "signatures": sigs
    }
    db_entry.append(certifiable_data)

    old_data = getEvent(did)
    if old_data is not None:
        print(old_data)
        for entry in old_data:
            db_entry.append(entry)

    subDb = dideryDB.open_db(DB_EVENT_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        txn.put(
            did.encode(),
            json.dumps(db_entry).encode()
        )

    return certifiable_data

def getAllEvents(offset=0, limit=10):
    """
        Get all events in a range between the offset and offset+limit

        :param offset: int starting point of the range
        :param limit: int maximum number of entries to return
        :return: dict
    """
    subDb = dideryDB.open_db(DB_EVENT_HISTORY_NAME)
    values = {"data": []}

    with dideryDB.begin(db=subDb, write=False) as txn:
        cursor = txn.cursor()

        count = 0
        for key, value in cursor:
            if count >= limit+offset:
                break

            if offset < count+1:
                values["data"].append(value)

            count += 1

    return values


def deleteEvent(did):
    """
        Find and delete a key rotation history matching the supplied did.

    :param did: string
        W3C did identifier for history object
    :return: boolean
    """
    subDb = dideryDB.open_db(DB_EVENT_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        status = txn.delete(did.encode())

        return status

def historyCount():
    """
        Gets a count of the number of entries in the table

        :return: int count
    """
    subDb = dideryDB.open_db(DB_KEY_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        return txn.stat(subDb)['entries']


def saveHistory(did, data, sigs):
    """
        Store a rotation history and signatures

        :param did: string
            W3C did string
        :param data: dict
            A dict containing the rotation history and signatures

    """
    certifiable_data = {
        "history": data,
        "signatures": sigs
    }
    subDb = dideryDB.open_db(DB_KEY_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        txn.put(
            did.encode(),
            json.dumps(certifiable_data).encode()
        )

    return certifiable_data


def getHistory(did):
    """
        Find and return a key rotation history matching the supplied did.

        :param did: string
            W3C did identifier for history object
        :return: dict
    """
    subDb = dideryDB.open_db(DB_KEY_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        raw_data = txn.get(did.encode())

        if raw_data is None:
            return None

        return json.loads(raw_data)


def getAllHistories(offset=0, limit=10):
    """
        Get all rotation histories in a range between the offset and offset+limit

        :param offset: int starting point of the range
        :param limit: int maximum number of entries to return
        :return: dict
    """
    subDb = dideryDB.open_db(DB_KEY_HISTORY_NAME)
    values = {"data": []}

    with dideryDB.begin(db=subDb, write=False) as txn:
        cursor = txn.cursor()

        count = 0
        for key, value in cursor:
            if count >= limit+offset:
                break

            if offset < count+1:
                values["data"].append(json.loads(value))

            count += 1

    return values


def deleteHistory(did):
    """
        Find and delete a key rotation history matching the supplied did.

    :param did: string
        W3C did identifier for history object
    :return: boolean
    """
    subDb = dideryDB.open_db(DB_KEY_HISTORY_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        status = txn.delete(did.encode())

        return status


def otpBlobCount():
    """
        Gets a count of the number of entries in the table

        :return: int count
    """
    subDb = dideryDB.open_db(DB_OTP_BLOB_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        return txn.stat(subDb)['entries']


def saveOtpBlob(did, data, sigs):
    """
        Store a otp encrypted key and signatures

        :param did: string
            W3C did string
        :param data: dict
            A dict containing the otp encrypted key and signatures

    """
    certifiable_data = {
        "otp_data": data,
        "signatures": sigs
    }
    subDb = dideryDB.open_db(DB_OTP_BLOB_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        txn.put(
            did.encode(),
            json.dumps(certifiable_data).encode()
        )

    return certifiable_data


def getOtpBlob(did):
    """
        Find and return an otp encrypted key matching the supplied did.

        :param did: string
            W3C did identifier for history object
        :return: dict
    """
    subDb = dideryDB.open_db(DB_OTP_BLOB_NAME)

    with dideryDB.begin(db=subDb, write=False) as txn:
        raw_data = txn.get(did.encode())

        if raw_data is None:
            return None

        return json.loads(raw_data)


def getAllOtpBlobs(offset=0, limit=10):
    """
            Get all otp encrypted keys in a range between the offset and offset+limit

            :param offset: int starting point of the range
            :param limit: int maximum number of entries to return
            :return: dict
        """
    subDb = dideryDB.open_db(DB_OTP_BLOB_NAME)
    values = {"data": []}

    with dideryDB.begin(db=subDb, write=False) as txn:
        cursor = txn.cursor()

        count = 0
        for key, value in cursor:
            if count >= limit + offset:
                break

            if offset < count + 1:
                values["data"].append(json.loads(value))

            count += 1

    return values


def deleteOtpBlob(did):
    """
        Find and delete a otp encrypted blob matching the supplied did.

    :param did: string
        W3C did identifier for history object
    :return: boolean
    """
    subDb = dideryDB.open_db(DB_OTP_BLOB_NAME)

    with dideryDB.begin(db=subDb, write=True) as txn:
        status = txn.delete(did.encode())

        return status


def loadTestData(name, data):
    pass



# -*- encoding: utf-8 -*-
"""
Keeping Module

"""
from __future__ import generator_stop

import os
import stat
import shutil
from collections import OrderedDict as ODict, deque
import enum
import binascii

try:
    import simplejson as json
except ImportError:
    import json

try:
    import msgpack
except ImportError:
    pass

import libnacl

from ioflo.aid.sixing import *
from ioflo.aid import filing
from ioflo.aid import getConsole

from ..help.helping import setupTmpBaseDir, makeDid

console = getConsole()

KEEP_DIR_PATH = "/var/bluepea/keep"  # default
ALT_KEEP_DIR_PATH = os.path.join('~', '.indigo/bluepea/keep')

gKeepDirPath = None  # key directory location has not been set up yet
gKeeper = None  # Keeper instance has not been set up yet

class Keeper(object):
    """
    Keeper Class to manage or hold cryptographic Keys
    """
    KeepDir = os.path.join("/var", "bluepea", "keep", )
    AltKeepDir = os.path.join("~", ".indigo", "bluepea", "keep")
    Prefix = "key"
    Role = "server"
    Ext = "json" # default serialization type of json and msgpack
    Fields = ("seed", "sigkey", "verkey", "prikey", "pubkey")

    def __init__(self,
                 baseDirPath="",
                 prefix="",
                 role="",
                 ext="",
                 fields=None,
                 seed=None,
                 prikey=None,
                 **kwa):
        """
        Setup Keeper instance
        Create directories for saving associated data files
            keep/
                bluepea/
                    prefix.role.ext

        """
        if not baseDirPath:
            baseDirPath = self.KeepDir
        baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))

        if not os.path.exists(baseDirPath):
            try:
                os.makedirs(baseDirPath)
            except OSError as ex:
                baseDirPath = self.AltKeepDir
                baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
                if not os.path.exists(baseDirPath):
                    os.makedirs(baseDirPath)

        else:
            if not os.access(baseDirPath, os.R_OK | os.W_OK):
                baseDirPath = self.AltKeepDir
                baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
                if not os.path.exists(baseDirPath):
                    os.makedirs(baseDirPath)

        self.baseDirPath = baseDirPath

        self.prefix = prefix if prefix else self.Prefix
        self.role = role if role else self.Role
        self.ext = ext if ext else self.Ext

        self.filePath = os.path.join(self.baseDirPath,
                                     "{}.{}.{}".format(self.prefix,
                                                        self.role,
                                                        self.ext))

        if fields is None:
            fields = self.Fields
        self.fields = tuple(field for field in fields)

        for field in self.fields:
            if not hasattr(self, field):
                setattr(self, field, None)

        self.setupKeys(seed=seed, prikey=prikey)
        self.did = makeDid(self.verkey)  # create did

    def setupKeys(self, seed, prikey):
        """
        Override stored keys with seed or prikey as provided

        """
        self.loadKeys()
        if seed:
            self.seed = seed
            self.sigkey = None
            self.verkey = None

        if prikey:
            self.prikey = prikey
            self.pubkey = None

        self.refreshKeys()
        self.dumpKeys()

    def refreshKeys(self):
        """
        Refreshes keys as needed so valid set of keys
        """
        if not self.seed: # no signing key seed so create new one
            self.seed = libnacl.randombytes(libnacl.crypto_sign_SEEDBYTES)

        if not self.sigkey or not self.verkey:  # ensure also sigkey and verkey
            self.verkey, self.sigkey = libnacl.crypto_sign_seed_keypair(self.seed)

        if not self.prikey: # no private encryption key so create new pair
            self.pubkey, self.prikey = libnacl.crypto_box_keypair()

        if not self.pubkey: # no public decryption key so create one from prikey
            self.pubkey = libnacl.crypto_scalarmult_base(self.prikey)

    def restoreKeys(self):
        """
        Reloads and refreshes as needed  and redumps  so valid set of keys
        """
        self.loadKeys()
        self.refreshKeys()
        self.dumpKeys()

    @staticmethod
    def dump(data, filePath):
        """
        Write data as as type self.ext to filepath. json or msgpack
        Sets permissions to only allow user to read and write to file
        """
        if " " in filePath:
            raise IOError("Invalid filepath '{0}' "
                                    "contains space".format(filePath))

        perm_other = stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
        perm_group = stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP
        cumask = os.umask(perm_other | perm_group)  # save old into cumask and set new

        root, ext = os.path.splitext(filePath)
        if ext == '.json':
            with filing.ocfn(filePath, "w+") as f:
                json.dump(data, f, indent=2)
                f.flush()
                os.fsync(f.fileno())
        elif ext == '.msgpack':
            if not msgpack:
                raise IOError("Invalid filepath ext '{0}' "
                            "needs msgpack installed".format(filePath))
            with filing.ocfn(filePath, "w+b", binary=True) as f:
                msgpack.dump(data, f)
                f.flush()
                os.fsync(f.fileno())
        else:
            raise IOError("Invalid filepath ext '{0}' "
                        "not '.json' or '.msgpack'".format(filePath))

        os.umask(cumask)  # restore old

    @staticmethod
    def load(filePath):
        """
        Return data read from filepath as converted json or msgpack
        Otherwise return None
        """
        try:
            root, ext = os.path.splitext(filePath)
            if ext == '.json':
                with filing.ocfn(filePath, "r") as f:
                    it = json.load(f, object_pairs_hook=ODict)
            elif ext == '.msgpack':
                if not msgpack:
                    raise IOError("Invalid filepath ext '{0}' "
                                "needs msgpack installed".format(filePath))
                with filing.ocfn(filePath, "rb", binary=True) as f:
                    it = msgpack.load(f, object_pairs_hook=ODict)
            else:
                it = None
        except EOFError:
            return None
        except ValueError:
            return None
        return it

    @staticmethod
    def loadAllRoles(dirPath="", prefix="key", role=""):
        """
        Load and Return the keys dict indexed by role for all key data files with
        prefix in  directory at dirPath  both .json and .msgpack file extensions
        are supported

        If role is not empty then loads last keyfile that matches both prefix and role

        key files names of form:
        prefix.role.json
        prefix.role.msgpack

        (when prefix is the default "key" )
        key.server.json
        key.server.msgpack

        key fields in keyfiles should be in:
        ('seed', 'sigkey', 'verkey', 'prikey', 'pubkey')

        values are unicode hex format

        """
        roles = ODict()
        for fileName in os.listdir(dirPath):  # filenames without directory
            filePath = os.path.join(dirPath, fileName)  # need full path for isfile
            if not os.path.isfile(filePath):
                continue
            root, ext = os.path.splitext(fileName)
            if ext not in ['.json', '.msgpack']:
                continue
            pre, sep, rol = root.partition('.')
            if not rol or pre != prefix or (role and rol != role):
                continue
            roles[rol] = loadKeys(filePath)
        return roles

    def clearBaseDir(self):
        """
        Clear the base directory
        """
        if os.path.exists(self.baseDirPath):
            shutil.rmtree(self.baseDirPath)  # shutil.rmtree

    def verifyKeyFields(self, keys, fields=None):
        """
        Returns True if the fields in fields match the fields in keys
        """
        fields = fields if fields else self.fields
        return (set(fields) == set(keys.keys()))

    def dumpKeys(self):
        """
        Dump Keeper keys to file in .baseDirpath
        key values stored in unicode hex format
        """
        keys = ODict()
        for field in self.fields:
            if hasattr(self, field):
                keys[field] = binascii.hexlify(getattr(self, field)).decode('utf-8')
        self.dump(keys, self.filePath)

    def loadKeys(self):
        """
        Load and verify Keeper keys from .filePath if any
        """
        if not os.path.exists(self.filePath):
            return
        keys = self.load(self.filePath)
        for field in self.fields:
            if field in keys and hasattr(self, field):
                setattr(self, field, binascii.unhexlify(keys[field].encode('utf-8')))

# Utility Functions

def setupKeeper(baseDirPath=None, seed=None, prikey=None):
    """
    Setup  the module global gKeepDirPath and gKeeper using baseDirPath, seed,
    and prikey if provided otherwise use KEEP_DIR_PATH and restore new keys

    """
    global gKeepDirPath, gKeeper

    if not baseDirPath:
        baseDirPath = KEEP_DIR_PATH

    baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
    if not os.path.exists(baseDirPath):
        try:
            os.makedirs(baseDirPath)
        except OSError as ex:
            baseDirPath = ALT_KEEP_DIR_PATH
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)
    else:
        if not os.access(baseDirPath, os.R_OK | os.W_OK):
            baseDirPath = ALT_KEEP_DIR_PATH
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)

    gKeepDirPath = baseDirPath  # set global

    gKeeper = Keeper(baseDirPath=gKeepDirPath, seed=seed, prikey=prikey)

    return gKeeper

def setupTestKeeper():
    """
    Return gKeeper resulting from baseDirpath in temporary directory
    and then setupKeeper
    """
    seed = (b'\x0c\xaa\xc9\xc6G\x11\xf6nn\xd7\x1b7\xdc^i\xc5\x12O\xe9>\xe1$F\xe1'
            b'\xa4z\xd4\xb6P\xdd\x86\x1d')

    prikey = (b'\xd9\xc8<$\x03\xb9%\x03c\xb3*6g\xa7m\xd8\x8d\x08j\xd4^4\x88\xcac\xba\xd1\xe9'
              b'\xd9\xe6\x99%')

    baseDirPath = setupTmpBaseDir()
    baseDirPath = os.path.join(baseDirPath, "bluepea/keep")
    os.makedirs(baseDirPath)
    return setupKeeper(baseDirPath=baseDirPath, seed=seed, prikey=prikey)

def setupKeep(baseDirPath=None):
    """
    Setup  the module global gKeepDirPath using baseDirPath
    if provided otherwise use KEEP_DIR_PATH

    """
    global gKeepDirPath

    if not baseDirPath:
        baseDirPath = KEEP_DIR_PATH

    baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
    if not os.path.exists(baseDirPath):
        try:
            os.makedirs(baseDirPath)
        except OSError as ex:
            baseDirPath = ALT_DATABASE_DIR_PATH
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)
    else:
        if not os.access(baseDirPath, os.R_OK | os.W_OK):
            baseDirPath = ALT_DATABASE_DIR_PATH
            baseDirPath = os.path.abspath(os.path.expanduser(baseDirPath))
            if not os.path.exists(baseDirPath):
                os.makedirs(baseDirPath)

    gKeepDirPath = baseDirPath  # set global

    return gKeepDirPath

def setupTestKeep():
    """
    Return gKeepDirPath resulting from baseDirpath in temporary directory
    and then setupKeep
    """
    baseDirPath = setupTmpBaseDir()
    baseDirPath = os.path.join(baseDirPath, "bluepea/keep")
    os.makedirs(baseDirPath)
    return setupKeep(baseDirPath=baseDirPath)

def dumpKeys(data, filepath):
    """
    Write key data as as type self.ext to filepath. json or msgpack
    Sets permissions to only allow user to read and write to file
    """
    if " " in filepath:
        raise IOError("Invalid filepath '{0}' "
                                "contains space".format(filepath))

    perm_other = stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
    perm_group = stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP
    cumask = os.umask(perm_other | perm_group)  # save old into cumask and set new

    root, ext = os.path.splitext(filepath)
    if ext == '.json':
        with filing.ocfn(filepath, "w+") as f:
            json.dump(data, f, indent=2)
            f.flush()
            os.fsync(f.fileno())
    elif ext == '.msgpack':
        if not msgpack:
            raise IOError("Invalid filepath ext '{0}' "
                        "needs msgpack installed".format(filepath))
        with filing.ocfn(filepath, "w+b", binary=True) as f:
            msgpack.dump(data, f)
            f.flush()
            os.fsync(f.fileno())
    else:
        raise IOError("Invalid filepath ext '{0}' "
                    "not '.json' or '.msgpack'".format(filepath))

    os.umask(cumask)  # restore old


def loadKeys(filepath):
    """
    Return key data read from filepath as converted json
    Otherwise return None
    """
    try:
        root, ext = os.path.splitext(filepath)
        if ext == '.json':
            with filing.ocfn(filepath, "r") as f:
                it = json.load(f, object_pairs_hook=ODict)
        elif ext == '.msgpack':
            if not msgpack:
                raise IOError("Invalid filepath ext '{0}' "
                            "needs msgpack installed".format(filepath))
            with filing.ocfn(filepath, "rb", binary=True) as f:
                it = msgpack.load(f, object_pairs_hook=ODict)
        else:
            it = None
    except EOFError:
        return None
    except ValueError:
        return None
    return it

def loadAllKeyRoles(dirpath, prefix="key", role=""):
    """
    Load and Return the keys dict indexed by role for all key data files with
    prefix in  directory at dirpath  both .json and .msgpack file extensions
    are supported

    If role is not empty then loads last keyfile that matches both prefix and role

    key files names of form:
    prefix.role.json
    prefix.role.msgpack

    (when prefix is the default "key" )
    key.server.json
    key.server.msgpack

    key fields in keyfiles should be in:
    ('seed', 'sigkey', 'verkey', 'prikey', 'pubkey')

    values are bytes of binary key value

    """
    roles = ODict()
    for filename in os.listdir(dirpath):  # filenames without directory
        filepath = os.path.join(dirpath, filename)  # need full path for isfile
        if not os.path.isfile(filepath):
            continue
        root, ext = os.path.splitext(filename)
        if ext not in ['.json', '.msgpack']:
            continue
        pre, sep, rol = root.partition('.')
        if not rol or pre != prefix or (role and rol != role):
            continue
        roles[rol] = loadKeys(filepath)
    return roles



import falcon
import arrow
try:
    import simplejson as json
except ImportError:
    import json

from ..help import helping
from .. import didering
from ..db import dbing as db


tempDB = {}


def basicValidation(req, resp, resource, params):
    """
        Do common validation tasks and prepare
        body of request for processing.
        :param req: Request object
    """
    raw = helping.parseReqBody(req)
    body = req.body

    required = ["id", "blob", "changed"]
    helping.validateRequiredFields(required, body)

    signature = req.get_header("Signature", required=True)
    sigs = helping.parseSignatureHeader(signature)
    req.signatures = sigs

    if len(sigs) == 0:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Empty Signature header.')

    sig = sigs.get('signer')  # str not bytes
    if not sig:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing "signer" tag and signature.')

    if body['id'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'id field cannot be empty.')

    if body['blob'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'blob field cannot be empty.')

    if body['changed'] == "":
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'changed field cannot be empty.')

    try:
        didKey = helping.extractDidParts(body['id'])
    except ValueError as ex:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               "Invalid did format. {}".format(str(ex)))

    return raw, sig, didKey


def validatePost(req, resp, resource, params):
    """
    Validate incoming POST requests.
    :param req: Request object
    :param resp: Response object
    :param resource: OtpBlob object
    :param params: dict of url params
    """
    raw, sig, didKey = basicValidation(req, resp, resource, params)

    try:
        helping.validateSignedResource(sig, raw, didKey)
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request body and signature. {}.'.format(ex))


def validatePut(req, resp, resource, params):
    """
    Validate incoming PUT requests.
    :param req: Request object
    :param resp: Response object
    :param resource: OtpBlob object
    :param params: dict of url params
    """

    if 'did' not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'DID value missing from url.')

    raw, sig, didKey = basicValidation(req, resp, resource, params)

    # Prevent did data from being clobbered
    if params['did'] != req.body['id']:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'Url did must match id field did.')

    try:
        helping.validateSignedResource(sig, raw, didKey)
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request body and signature. {}.'.format(ex))


def validateDelete(req, resp, resource, params):
    if 'did' not in params:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'DID value missing from url.')

    raw = helping.parseReqBody(req)
    signature = req.get_header("Signature", required=True)
    sigs = helping.parseSignatureHeader(signature)
    req.signatures = sigs
    body = req.body

    if len(sigs) == 0:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Empty Signature header.')

    signer = sigs.get('signer')  # str not bytes
    if not signer:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Signature header missing signature for "signer".')

    # Prevent did data from being clobbered
    if params['did'] != body['id']:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Validation Error',
                               'Url did must match id field did.')

    otp = db.getOtpBlob(params['did'])
    req.otp = otp
    if otp is None:
        raise falcon.HTTPError(falcon.HTTP_404)

    vk = helping.extractDidParts(otp["otp_data"]['id'])

    try:
        helping.validateSignedResource(signer, raw, vk)
    except didering.ValidationError as ex:
        raise falcon.HTTPError(falcon.HTTP_401,
                               'Authorization Error',
                               'Could not validate the request signature for signer field. {}.'.format(ex))


class OtpBlob:
    def __init__(self, store=None):
        """
        :param store: Store
            store is reference to ioflo data store
        """
        self.store = store

    """
    For manual testing of the endpoint:
        http localhost:8000/blob/did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=
        http localhost:8000/blob
    """
    @falcon.before(helping.parseQString)
    def on_get(self, req, resp, did=None):
        """
        Handle and respond to incoming GET request.
        :param req: Request object
        :param resp: Response object
        :param did: string
            URL parameter specifying a otp encrypted key
        """
        offset = req.offset
        limit = req.limit

        count = db.otpBlobCount()

        if did is not None:
            body = db.getOtpBlob(did)
            if body is None:
                raise falcon.HTTPError(falcon.HTTP_404)
        else:
            if offset >= count:
                resp.body = json.dumps({}, ensure_ascii=False)
                return

            body = db.getAllOtpBlobs(offset, limit)

            resp.append_header('X-Total-Count', count)

        resp.body = json.dumps(body, ensure_ascii=False)


    """
        For manual testing of the endpoint:
        http POST localhost:8000/blob id="did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=" blob="AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
    """
    @falcon.before(validatePost)
    def on_post(self, req, resp):
        """
        Handle and respond to incoming POST request.
        :param req: Request object
        :param resp: Response object
        """
        result_json = req.body
        sigs = req.signatures
        did = result_json['id']

        if db.getOtpBlob(did) is not None:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Resource Already Exists',
                                   'Resource with did "{}" already exists. Use PUT request.'.format(result_json['id']))

        # TODO: review signature validation for any holes
        response_json = db.saveOtpBlob(did, result_json, sigs)

        resp.body = json.dumps(response_json, ensure_ascii=False)
        resp.status = falcon.HTTP_201

    """
        For manual testing of the endpoint:
        http PUT localhost:8000/blob/did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE= id="did:dad:Qt27fThWoNZsa88VrTkep6H-4HA8tr54sHON1vWl6FE=" blob="AeYbsHot0pmdWAcgTo5sD8iAuSQAfnH5U6wiIGpVNJQQoYKBYrPPxAoIc1i5SHCIDS8KFFgf8i0tDq8XGizaCgo9yjuKHHNJZFi0QD9K6Vpt6fP0XgXlj8z_4D-7s3CcYmuoWAh6NVtYaf_GWw_2sCrHBAA2mAEsml3thLmu50Dw"
    """
    @falcon.before(validatePut)
    def on_put(self, req, resp, did):
        """
        Handle and respond to incoming PUT request.
        :param req: Request object
        :param resp: Response object
        :param did: did string
        """
        result_json = req.body
        sigs = req.signatures

        resource = db.getOtpBlob(did)

        if resource is None:
            raise falcon.HTTPError(falcon.HTTP_404)

        current = arrow.get(resource['otp_data']['changed'])
        update = arrow.get(result_json['changed'])
        if current >= update:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Validation Error',
                                   '"changed" field not later than previous update.')

        # TODO: review signature validation for any holes
        response_json = db.saveOtpBlob(did, result_json, sigs)

        resp.body = json.dumps(response_json, ensure_ascii=False)

    @falcon.before(validateDelete)
    def on_delete(self, req, resp, did):
        """
            Handle and respond to incoming PUT request.
            :param req: Request object
            :param resp: Response object
            :param did: decentralized identifier
        """
        resource = req.otp

        success = db.deleteOtpBlob(did)

        if not success:
            raise falcon.HTTPError(falcon.HTTP_500,
                                   'Deletion Error',
                                   'Error while attempting to delete the resource.')

        resp.body = json.dumps({"deleted": resource}, ensure_ascii=False)



# ================================================== #
#                     DASHBOARD                      #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/14/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class Searcher:
    """
    Class to search for a certain string in any dict object.
    """
    def __init__(self):
        """
        Initializes Searcher object.
        """
        self.searchTerm = None
        self.caseSensitive = False

    # ============================================== #

    def setSearch(self, term):
        """
        Sets search term. If term is surrounded by quotes, removes
        them and makes search case sensitive. Otherwise, search is
        not case sensitive.

            Parameters::
            term - Base search string
        """
        self.searchTerm = term or ""
        self.caseSensitive = self.searchTerm.startswith('"') and self.searchTerm.endswith('"')
        if self.caseSensitive:
            self.searchTerm = self.searchTerm[1:-1]
        else:
            self.searchTerm = self.searchTerm.lower()

    # ============================================== #

    def _checkPrimitive(self, item):
        """
        Checks for search term in provided string.
        """
        if isinstance(item, str):
            if not self.caseSensitive:
                item = item.lower()
            return self.searchTerm in item
        return False

    # ============================================== #

    def _checkAny(self, value):
        """
        Checks for search term in any provided dict, list, or primitive type.
        """
        if isinstance(value, dict) or isinstance(value, Object):
            return self.search(value)
        elif isinstance(value, list):
            for item in value:
                if self._checkAny(item):
                    return True
            return False
        else:
            return self._checkPrimitive(value)

    # ============================================== #

    def search(self, obj: dict):
        """
        Returns True if obj recursively contains search term string in any field.
        """
        __pragma__("jsiter")
        for key in obj:
            if key.startswith("_"):
                continue

            value = obj[key]
            if self._checkAny(value):
                return True
        return False
        __pragma__("nojsiter")

# ================================================== #
#                        EOF                         #
# ================================================== #


from collections import OrderedDict as ODict
import falcon
import libnacl
import base64
import tempfile
import os
import shutil

try:
    import simplejson as json
except ImportError:
    import json

from .. import didering


def parseSignatureHeader(signature):
    """
    Returns ODict of fields and values parsed from signature
    which is the value portion of a Signature header

    Signature header has format:

        Signature: headervalue

    Headervalue:
        tag = "signature"

    or

        tag = "signature"; tag = "signature"  ...

    where tag is the name of a field in the body of the request whose value
    is a DID from which the public key for the signature can be obtained
    If the same tag appears multiple times then only the last occurrence is returned

    each signature value is a doubly quoted string that contains the actual signature
    in Base64 url safe format. By default the signatures are EdDSA (Ed25519)
    which are 88 characters long (with two trailing pad bytes) that represent
    64 byte EdDSA signatures

    An option tag name = "kind" with values "EdDSA"  "Ed25519" may be present
    that specifies the type of signature. All signatures within the header
    must be of the same kind.

    The two tag fields currently supported are "did" and "signer"


    """
    sigs = ODict()
    if signature:
        clauses = signature.split(";")
        for clause in clauses:
            clause = clause.strip()
            if not clause:
                continue
            try:
                tag, value = clause.split("=", maxsplit=1)
            except ValueError as ex:
                continue
            tag = tag.strip()
            if not tag:
                continue
            value = value.strip()
            if not value.startswith('"') or not value.endswith('"') or len(value) < 3:
                continue
            value = value[1:-1]
            value = value.strip()
            sigs[tag] = value

    return sigs


def validateSignedResource(signature, resource, verkey, method="dad"):
    """
    Returns dict of deserialized resource if signature verifies for resource given
    verification key verkey in base64 url safe unicode format
    Otherwise returns None


    signature is base64 url-file safe unicode string signature generated
        by signing bytes version of resource with privated signing key associated with
        public verification key referenced by key indexed signer field in resource

    resource is json encoded unicode string of resource record

    verkey is base64 url-file safe unicode string public verification key referenced
        by signer field in resource. This is looked up in database from signer's
        agent data resource

    method is the method string used to generate dids in the resource
    """

    try:
        try:
            rsrc = json.loads(resource, object_pairs_hook=ODict)
        except ValueError as ex:
            raise didering.ValidationError("Invalid JSON")  # invalid json

        ddid = rsrc["id"]

        try:  # correct did format  pre:method:keystr
            pre, meth, keystr = ddid.split(":")
        except ValueError as ex:
            raise didering.ValidationError("Invalid format did field")

        if pre != "did" or meth != method:
            raise didering.ValidationError("Invalid format did field") # did format bad

        if len(verkey) != 44:
            raise didering.ValidationError("Verkey invalid")  # invalid length for base64 encoded key

        if not verify64u(signature, resource, verkey):
            raise didering.ValidationError("Unverifiable signature")  # signature fails

    except didering.ValidationError:
        raise

    except Exception as ex:  # unknown problem
        raise didering.ValidationError("Unexpected error")

    return rsrc


def parseReqBody(req):
    """
    :param req: Falcon Request object
    """
    try:
        raw_json = req.stream.read()
    except Exception as ex:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Request Error',
                               'Error reading request body.')

    try:
        result_json = json.loads(raw_json, encoding='utf-8')
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Malformed JSON',
                               'Could not decode the request body. The '
                               'JSON was incorrect.')

    req.body = result_json
    return raw_json


def validateRequiredFields(required, resource):
    """
    Validate that all required fields are present in resource
    :param required: list of required fields
    :param resource: dict to be checked
    """
    for req in required:
        if req not in resource:
            raise falcon.HTTPError(falcon.HTTP_400,
                                   'Missing Required Field',
                                   'Request must contain {} field.'.format(req))


def keyToKey64u(key):
    """
    Convert and return bytes key to unicode base64 url-file safe version
    """
    return base64.urlsafe_b64encode(key).decode("utf-8")


def key64uToKey(key64u):
    """
    Convert and return unicode base64 url-file safe key64u to bytes key
    """
    return base64.urlsafe_b64decode(key64u.encode("utf-8"))


def makeDid(vk, method="dad"):
    """
    Create and return Did from bytes vk.
    vk is 32 byte verifier key from EdDSA (Ed25519) keypair
    """
    # convert verkey to jsonable unicode string of base64 url-file safe
    vk64u = base64.urlsafe_b64encode(vk).decode("utf-8")
    did = "did:{}:{}".format(method, vk64u)
    return did


def signResource(resource, sKey):
    sig = libnacl.crypto_sign(resource, sKey)
    sig = sig[:libnacl.crypto_sign_BYTES]

    return keyToKey64u(sig)


def verify(sig, msg, vk):
    """
    Returns True if signature sig of message msg is verified with
    verification key vk Otherwise False
    All of sig, msg, vk are bytes
    """
    try:
        result = libnacl.crypto_sign_open(sig + msg, vk)
    except Exception as ex:
        return False
    return True if result else False


def verify64u(signature, message, verkey):
    """
    Returns True if signature is valid for message with respect to verification
    key verkey

    signature and verkey are encoded as unicode base64 url-file strings
    and message is unicode string as would be the case for a json object

    """
    sig = key64uToKey(signature)
    vk = key64uToKey(verkey)
    # msg = message.encode("utf-8")
    return verify(sig, message, vk)


def extractDidParts(did, method="dad"):
    """
    Parses and returns keystr from did
    raises ValueError if fails parsing
    """
    try:  # correct did format  pre:method:keystr
        pre, meth, keystr = did.split(":")
    except ValueError as ex:
        raise ValueError("Invalid DID value")

    if pre != "did":
        raise ValueError("Invalid DID identifier")

    if meth != method:
        raise ValueError("Invalid DID method")

    return keystr


def genDidHistory(seed, changed="2000-01-01T00:00:00+00:00", signer=0, numSigners=3):
    # seed = libnacl.randombytes(libnacl.crypto_sign_SEEDBYTES)
    vk, sk = libnacl.crypto_sign_seed_keypair(seed)

    did = makeDid(vk)
    body = {
        "id": did,
        "changed": changed,
        "signer": signer,
        "signers": []
    }

    for i in range(0, numSigners):
        body['signers'].append(keyToKey64u(vk))

    return vk, sk, did, json.dumps(body, ensure_ascii=False).encode()


def parseQString(req, resp, resource, params):
    req.offset = 0
    req.limit = 10

    if req.query_string:
        queries = req.query_string.split('&')
        for query in queries:
            key, val = qStringValidation(query)
            if key == 'offset':
                req.offset = val
            if key == 'limit':
                req.limit = val


def qStringValidation(query):
    keyval = query.split('=')

    if len(keyval) != 2:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Malformed Query String',
                               'url query string missing value(s).')

    key = keyval[0]
    val = keyval[1]

    try:
        val = int(val)
    except ValueError as ex:
        raise falcon.HTTPError(falcon.HTTP_400,
                               'Malformed Query String',
                               'url query string value must be a number.')

    return key, val


def verifyPublicApiRequest(reqFunc, url, body, headers=None, exp_result=None, exp_status=None):
    SK = b"\xb3\xd0\xbdL]\xcc\x08\x90\xa5\xbd\xc6\xa1 '\x82\x9c\x18\xecf\xa6x\xe2]Ux\xa5c\x0f\xe2\x86*\xa04\xe7\xfaf\x08o\x18\xd6\xc5s\xfc+\xdc \xb4\xb4\xa6G\xcfZ\x96\x01\x1e%\x0f\x96\x8c\xfa-3J<"

    body = json.dumps(body, ensure_ascii=False).encode('utf-8')

    if headers is None:
        headers = {
            "Signature": 'signer="{0}"; rotation="{1}"'.format(signResource(body, SK),
                                                               signResource(body, SK))
        }

    response = reqFunc(url, body=body, headers=headers)

    print("status: {0},  content: {1}".format(response.status, response.content))
    if exp_result is not None:
        assert json.loads(response.content) == exp_result
    if exp_status is not None:
        assert response.status == exp_status


def verifyManagementApiRequest(reqFunc, url, body=None, exp_result=None, exp_status=None):

    response = reqFunc(url, body=json.dumps(body, ensure_ascii=False))

    print("status: {0},  content: {1}".format(response.status, response.content))
    if exp_result is not None:
        assert json.loads(response.content) == exp_result
    if exp_status is not None:
        assert response.status == exp_status


def setupTmpBaseDir():
    """
    Create temporary directory
    """
    return tempfile.mkdtemp(prefix="didery",  suffix="test", dir="/tmp")


def cleanupTmpBaseDir(baseDirPath):
    """
    Remove temporary root of baseDirPath
    Ascend tree to find temporary root directory
    """
    if os.path.exists(baseDirPath):
        shutil.rmtree(baseDirPath)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Basic setup file to enable pip install
See:
    https://pythonhosted.org/setuptools/
    https://bitbucket.org/pypa/setuptools


$ python setup.py register sdist upload

More secure to use twine to upload
$ pip3 install twine
$ python3 setup.py sdist
$ twine upload dist/toba-0.1.0.tar.gz

"""

from __future__ import generator_stop

import sys
import io
import os
import re

v = sys.version_info
if v < (3, 6):
    msg = "FAIL: Requires Python 3.6 or later, but setup.py was run using {}.{}.{}"
    print(msg.format(v.major, v.minor, v.micro))
    print("NOTE: Installation failed. Run setup.py using python3")
    sys.exit(1)


from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import relpath
from os.path import splitext

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_ext import build_ext

try:
    # Allow installing package without any Cython available. This
    # assumes you are going to include the .c files in your sdist.
    import Cython
except ImportError:
    Cython = None


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


# Enable code coverage for C code: we can't use CFLAGS=-coverage in tox.ini, since that may mess with compiling
# dependencies (e.g. numpy). Therefore we set SETUPPY_CFLAGS=-coverage in tox.ini and copy it to CFLAGS here (after
# deps have been safely installed).
if 'TOXENV' in os.environ and 'SETUPPY_CFLAGS' in os.environ:
    os.environ['CFLAGS'] = os.environ['SETUPPY_CFLAGS']


class optional_build_ext(build_ext):
    """Allow the building of C extensions to fail."""
    def run(self):
        try:
            build_ext.run(self)
        except Exception as e:
            self._unavailable(e)
            self.extensions = []  # avoid copying missing files (it would fail).

    def _unavailable(self, e):
        print('*' * 80)
        print('''WARNING:

    An optional code optimization (C extension) could not be compiled.

    Optimizations for this package will not be available!
        ''')

        print('CAUSE:')
        print('')
        print('    ' + repr(e))
        print('*' * 80)


setup(
    name='didery',
    version='0.0.3',
    license='Apache2',
    description='DIDery Key Management Server',
    long_description="Redundant persistent backup of key rotation events and otp encrypted private keys.",
    author='Nicholas Telfer, Brady Hammond, Michael Mendoza',
    author_email='nick.telfer@consensys.net',
    url='https://github.com/reputage/didery',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    package_data={
        'didery': ['static/main.html',
                   'static/css/*.css',
                   'static/fonts/Raleway/*.ttf',
                   'static/node_modules/mithril/mithril.min.js',
                   'static/node_modules/jquery/dist/jquery.min.js',
                   'static/node_modules/semantic-ui/dist/semantic.min.css',
                   'static/node_modules/semantic-ui/dist/semantic.min.js',
                   'static/node_modules/semantic-ui/dist/themes/default/assets/fonts/*.woff2',
                   'static/node_modules/semantic-ui/dist/themes/default/assets/fonts/*.woff',
                   'static/node_modules/semantic-ui/dist/themes/default/assets/fonts/*.ttf',
                   'static/transcrypt/__javascript__/main.js',
                   'flo/*.flo'
                   ]
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Utilities',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'click', 'falcon>=1.2', 'ioflo>=1.6.8', 'libnacl>=1.5.1',
        'simplejson>=3.11.1', 'pytest-falcon>=0.4.2', 'arrow>=0.10.0',
        'transcrypt<=3.6.101', 'lmdb',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    setup_requires=[
        'cython',
    ] if Cython else [],
    entry_points={
        'console_scripts': [
            'dideryd = didery.cli:main',
        ]
    },
    cmdclass={'build_ext': optional_build_ext},
    ext_modules=[
        Extension(
            splitext(relpath(path, 'src').replace(os.sep, '.'))[0],
            sources=[path],
            include_dirs=[dirname(path)]
        )
        for root, _, _ in os.walk('src')
        for path in glob(join(root, '*.pyx' if Cython else '*.c'))
    ],
)



# ================================================== #
#                       TABLES                       #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/08/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

import didery.static.transcrypt.server as server
import didery.static.transcrypt.components.fields as field

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class Table:
    """
    Class for table, headers, and display data..
    """
    no_results_text = "No results found."

    def __init__(self, fields):
        """
        Initializes Table object.

            Parameters:
            fields - Table fields objects
        """
        self.max_size = 1000
        self.fields = fields
        self.data = []
        self._shownData = []
        self.view = {
            "view": self._view
        }

        self._selected = None
        self.detailSelected = ""

        self.filter = None
        self.sortField = None
        self.reversed = False

        self.total = 0
        self.shown = 0

    # ============================================== #

    def _stringify(self, obj):
        """
        Converts provided json-like object to user-friendly string.

            Parameters:
            obj - JSON like object
        """
        def replacer(key, value):
            if key.startswith("_"):
                return
            return value
        return JSON.stringify(obj, replacer, 2)

    # ============================================== #

    def _limitText(self):
        """
        Limits displayed results.
        """
        return "Limited to {} results.".format(self.max_size)

    # ============================================== #

    def _selectRow(self, event, obj):
        """
        Deselects any previously selected row and selects row
        specified in event.

            Parameters:
            event - Specified event
            obj - Row data object
        """
        if self._selected is not None:
            del self._selected._selected

            if self._selected._uid == obj._uid:
                self._selected = None
                self.detailSelected = ""
                return

        self._selected = obj
        obj._selected = True
        self.detailSelected = self._stringify(obj)

    # ============================================== #

    def refresh(self):
        """
        Refreshes data from server and returns a promise that resolves
        when finished.
        """
        self._setData([])
        p = __new__(Promise(lambda resolve: resolve()))
        return Promise

    # ============================================== #

    def clear(self):
        """
        Removes memory of all current data.
        """
        self.total = 0
        server.clearArray(self.data)

    # ============================================== #

    def _makeDummyData(self, count):
        """
        Creates test data.

            Parameters:
            count - Integer
        """
        data = []
        for i in range(count):
            obj = {}
            for field in self.fields:
                obj[field.name] = "test{0} {1}".format(i, field.name)
            data.append(obj)
        return data

    # ============================================== #

    __pragma__("kwargs")
    def _setData(self, data, clear=True):
        """
        Clears existing data and uses provided data instead.
        Adds "_uid" field to each piece of data, for internal tracking.

            Parameters:
            data - Input data
            clear - Boolean
        """
        if clear:
            self.clear()
        for datum in data:
            datum._uid = self.total
            self.data.append(datum)
            self.total += 1
        self._processData()
    __pragma__("nokwargs")

    # ============================================== #

    def setFilter(self, func):
        """
        Sets table filter function.

            Parameters:
            func - Filter function
        """
        if func != self.filter:
            self.filter = func
            self._processData()

    # ============================================== #

    def setSort(self, field):
        """
        Sets sort on given field. If same as currently-sorted field,
        reverse sort on said field.

            Parameters:
            field - Field to sort by
        """
        if self.sortField == field:
            self.reversed = not self.reversed
        else:
            self.reversed = False
            self.sortField = field

        self._sortData()

    # ============================================== #

    def _sortData(self):
        """
        Sorts table data.
        """
        if self.sortField is None:
            return

        self._shownData.sort(key=lambda obj: self._getField(obj, self.sortField), reverse=self.reversed)

    # ============================================== #

    def _processData(self):
        """
        Processes data, determines which items to show, and
        puts items into sorted list.
        """
        server.clearArray(self._shownData)

        self.shown = 0
        for obj in self.data:
            if self.shown >= self.max_size:
                break
            if self.filter is not None:
                if not self.filter(obj):
                    continue

            self._shownData.append(obj)
            self.shown += 1

        self._sortData()

    # ============================================== #

    def _getField(self, obj, field):
        """
        Gets info from object matching given field.

            Parameters:
            obj - Data object
            field - Field/Key
        """
        return obj[field.name]

    # ============================================== #

    def _makeRow(self, obj):
        """
        Returns array of <td> vnodes representing a row.

            Parameters:
            obj - Data object
        """
        return [field.view(self._getField(obj, field)) for field in self.fields]

    # ============================================== #

    def _view(self):
        """
        Returns table markup.
        """
        headers = []
        for field in self.fields:
            def makeScope(f):
                return lambda event: self.setSort(f)
            if field == self.sortField:
                if self.reversed:
                    icon = m("i.angle.down.icon")
                else:
                    icon = m("i.angle.up.icon")
                header = m("th.ui.right.labeled.icon", {"onclick": makeScope(field)},
                           icon,
                           field.title)
            else:
                header = m("th", {"onclick": makeScope(field)}, field.title)

            headers.append(header)

        rows = []
        for obj in self._shownData:
            row = self._makeRow(obj)

            def makeScope(o):
                return lambda event: self._selectRow(event, o)
            if obj._selected:
                rows.append(m("tr.active", {"onclick": makeScope(obj)}, row))
            else:
                rows.append(m("tr", {"onclick": makeScope(obj)}, row))

        if self.shown >= self.max_size:
            rows.append(m("tr", m("td", self._limitText())))

        if not self.shown:
            rows.append(m("tr", m("td", self.no_results_text)))

        return m("table", {"class": "ui selectable celled unstackable single line left aligned table"},
                 m("thead",
                   m("tr", {"class": "center aligned"}, headers)
                   ),
                 m("tbody",
                   rows
                   )
                 )

# ================================================== #

class ErrorsTable(Table):
    """
    Class for errors table.
    """
    def __init__(self):
        """
        Initializes ErrorsTable object. Sets up error table fields.
        """
        fields = [
            field.FillField("Title"),
            field.FillField("Message"),
            field.DateField("Time")
        ]
        super().__init__(fields)

    # ============================================== #

    def refresh(self):
        """
        Refreshes table data.
        """
        self.clear()
        errors = server.manager.errors
        return errors.refreshErrors().then(lambda: self._setData(errors.errors))

    # ============================================== #

    def _getField(self, obj, field):
        """
        Extracts data from json-like object.

            Parameters:
            obj - Data object
            field - Field/Key
        """
        if field.name == "title":
            return obj.title
        elif field.name == "message":
            return obj.msg
        elif field.name == "time":
            return obj.time

# ================================================== #

class RelaysTable(Table):
    """
    Class for relays table.
    """
    def __init__(self):
        """
        Initializes RelaysTable object. Sets up relays table fields.
        """
        fields = [
            field.FillField("Host"),
            field.FillField("Port"),
            field.FillField("Name"),
            field.FillField("Main"),
            field.IDField("UID"),
            field.FillField("Status")
        ]
        super().__init__(fields)

    # ============================================== #

    def refresh(self):
        """
        Refreshes table data.
        """
        self.clear()
        relays = server.manager.relays
        return relays.refreshRelays().then(lambda: self._setData(relays.relays))

    # ============================================== #

    def _getField(self, obj, field):
        """
        Extracts data from json-like object.

            Parameters:
            obj - Data object
            field - Field/Key
        """
        if field.name == "host":
            return obj["host_address"]
        elif field.name == "port":
            return obj["port"]
        elif field.name == "name":
            return obj["name"]
        elif field.name == "main":
            return obj["main"]
        elif field.name == "uid":
            return obj["uid"]
        elif field.name == "status":
            return obj["status"]

# ================================================== #

class BlobsTable(Table):
    """
    Class for blobs table.
    """
    def __init__(self):
        """
        Initializes BlobsTable object. Sets up blobs table fields.
        """
        fields = [
            field.DIDField("DID"),
            field.FillField("Blob")
        ]
        super().__init__(fields)

    # ============================================== #

    def refresh(self):
        """
        Refreshes table data.
        """
        self.clear()
        blobs = server.manager.otpBlobs
        return blobs.refreshBlobs().then(lambda: self._setData(blobs.blobs))

    # ============================================== #

    def _getField(self, obj, field):
        """
        Extracts data from json-like object.

            Parameters:
            obj - Data object
            field - Field/Key
        """

        if field.name == "did":
            return obj.id
        elif field.name == "blob":
            return obj.blob

# ================================================== #

class HistoryTable(Table):
    """
    Class for history table.
    """
    def __init__(self):
        """
        Initializes HistoryTable object. Sets up history table fields.
        """
        fields = [
            field.DIDField("DID"),
            field.DateField("Changed"),
            field.FillField("Signer"),
            field.FillField("Signers"),
            field.FillField("Signatures")
        ]
        super().__init__(fields)

    # ============================================== #

    def refresh(self):
        """
        Refreshes table data.
        """
        self.clear()
        history = server.manager.history
        return history.refreshHistory().then(lambda: self._setData(history.history))

    # ============================================== #

    def _getField(self, obj, field):
        """
        Extracts data from json-like object.

            Parameters:
            obj - Data object
            field - Field/Key
        """

        if field.name == "did":
            return obj.history.id
        elif field.name == "changed":
            return obj.history.changed
        elif field.name == "signer":
            return obj.history.signer
        elif field.name == "signers":
            signers = ""
            for signer in obj.history.signers:
                signers += signer + ", "
            signers = signers[:-2]
            return signers
        elif field.name == "signatures":
            signatures = ""
            for key, value in dict(obj.signatures).items():
                signatures += value + ", "
            signatures = signatures[:-2]
            return signatures

# ================================================== #
#                        EOF                         #
# ================================================== #


# ================================================== #
#                     TABLED TAB                     #
# ================================================== #
# Author: Brady Hammond                              #
# Created: 05/08/2018                                #
# Last Edited:                                       #
# Last Edited By:                                    #
# ================================================== #
#                      IMPORTS                       #
# ================================================== #

import didery.static.transcrypt.components.tab as tab

# ================================================== #
#                  CLASS DEFINITIONS                 #
# ================================================== #

class TabledTab(tab.Tab):
    """
    Base class for tabs in the dashboard interface. Uses table and details view.
    """
    def __init__(self):
        """
        Initialize TabledTab object.
        """
        super().__init__()
        self.table = None
        self.setup_table()
        self.copiedDetails = ""
        self._detailsId = self.DataTab + "DetailsCodeBlock"
        self._copiedId = self.DataTab + "CopiedCodeBlock"
        self._copyButtonId = self.DataTab + "CopyButton"
        self._clearButtonId = self.DataTab + "ClearButton"

    # ============================================== #

    def setup_table(self):
        """
        Sets up table object.
        """
        self.table = Table([])

    # ============================================== #

    def _copyDetails(self):
        self.copiedDetails = self.table.detailSelected

    # ============================================== #

    def _getRows(self):
        return jQuery("[data-tab='{0}'].tab table > tbody > tr".format(self.DataTab))

    # ============================================== #

    def _getLabel(self):
        return jQuery(".menu a[data-tab='{0}'] .ui.label".format(self.DataTab))

    # ============================================== #

    def _clearCopy(self):
        self.copiedDetails = ""

    # ============================================== #

    def menu_item(self):
        return m(self._menu, self._menu_attrs,
                 m("div.menu-item-text", self.Name),
                 m(self.Icon),
                 m("div.ui.label.small.menu-item-number", "{0}/{1}".format(self.table.shown, self.table.total))
                 )

    # ============================================== #

    def main_view(self):
        return m("div",
                 m("div.table-container", m(self.table.view)),
                 m("div.ui.hidden.divider"),
                 m("div.ui.two.cards", {"style": "height: 45%;"},
                   m("div.ui.card",
                     m("div.content.small-header",
                       m("div.header",
                         m("span", "Details"),
                         m("span.ui.mini.right.floated.button", {"onclick": self._copyDetails, "id": self._copyButtonId},
                           "Copy")
                         )
                       ),
                     m("pre.content.code-block", {"id": self._detailsId},
                       self.table.detailSelected
                       )
                     ),
                   m("div.ui.card",
                     m("div.content.small-header",
                       m("div.header",
                         m("span", "Copied"),
                         m("span.ui.mini.right.floated.button", {"onclick": self._clearCopy, "id": self._clearButtonId},
                           "Clear")
                         )
                       ),
                     m("pre.content.code-block", {"id": self._copiedId},
                       self.copiedDetails
                       )
                     )
                   )
                 )

# ================================================== #
#                        EOF                         #
# ================================================== #


"""
controllers package

rest endpoints

"""
from __future__ import generator_stop

import importlib

_modules = ['errors', 'events', 'history', 'otp_blob', 'relays', 'static']

for m in _modules:
    importlib.import_module(".{0}".format(m), package='didery.controllers')



import falcon

from ioflo.aid import getConsole
from ioflo.aid import odict
from ioflo.aio.http import Valet
from ioflo.base import doify

from didery import routing
from didery.db import dbing

console = getConsole()

"""
Usage pattern

frame server
    do didery server open at enter
    do didery server service
    do didery server close at exit
"""


@doify('DideryServerOpen', ioinits=odict(
                                        valet="",
                                        test="",
                                        port=odict(ival=8080),
                                        db=odict(ival=""),
                                        ))
def dideryServerOpen(self):
    """
    Setup and open a rest server

    Ioinit attributes
        valet is Valet instance (wsgi server)
        port is server port

    Context: enter

    Example:
        do didery server open at enter
    """
    port = int(self.port.value)
    dbing.setupDbEnv(self.db.value, self.port.value)

    app = falcon.API(middleware=[routing.CORSMiddleware()])
    routing.loadEndPoints(app, store=self.store)

    self.valet.value = Valet(
                            port=port,
                            bufsize=131072,
                            wlog=None,
                            store=self.store,
                            app=app,
                            timeout=0.5,
                            )

    console.terse("IP Address {}\n".format(self.valet.value.servant.ha))

    result = self.valet.value.servant.reopen()

    if not result:
        console.terse("Error opening server '{0}' at '{1}'\n".format(
            self.valet.name,
            self.valet.value.servant.ha))
        return

    console.concise("Opened server '{0}' at '{1}'\n".format(
        self.valet.name,
        self.valet.value.servant.ha, ))


@doify('DideryServerService', ioinits=odict(valet=""))
def dideryServerService(self):
    """
    Service server given by valet

    Ioinit attributes:
        valet is a Valet instance

    Context: recur

    Example:
        do didery server service
    """
    if self.valet.value:
        self.valet.value.serviceAll()


@doify('DideryServerClose', ioinits=odict(valet="",))
def dideryServerClose(self):
    """
    Close server in valet

    Ioinit attributes:
        valet is a Valet instance

    Context: exit

    Example:
        do didery server close at exit
    """
    if self.valet.value:
        self.valet.value.servant.closeAll()

        console.concise("Closed server '{0}' at '{1}'\n".format(
                            self.valet.name,
                            self.valet.value.servant.eha))



class DideringError(Exception):
    """
    Base Class for didery exceptions

    To use   raise DideringError("Error: message")
    """


class ValidationError(DideringError):
    """
    Validation related errors
    Usage:
        raise ValidationError("error message")
    """
